def scale_x(inp, min_, max_):
    return (inp - min_) / (max_ - min_)


def scale_par(inp, min_, max_):
    return scale_x(inp, min_, max_)


def scale_y(inp, mean_, scale_):
    return (inp - mean_) / (max_ - min_)


def unscale_x(inp, min_, max_):
    return inp * (max_ - min_) + min_


def unscale_par(inp, min_, max_):
    return unscale_x(inp, min_, max_)


def unscale_y(inp, mean_, scale_):
    return inp * scale_ + mean_



import os
import random
import warnings

import numpy as np
import pandas as pd
from prettyPlot.plotting import *
from sklearn.preprocessing import OneHotEncoder

from bird.postprocess.SA_optimization.surrogate import (
    Surrogate_wrapper,
    tune_nn,
    tune_rbf,
    tune_rf,
)

warnings.filterwarnings("ignore")


def simulated_annealing_surrogate(
    surrogate: Surrogate_wrapper,
    dim: int = 12,
    max_iters: int = 1000,
    temp: float = 10.0,
    alpha: float = 0.95,
) -> tuple[np.ndarray, float, list[tuple], list[np.ndarray]]:
    """
    Runs the Simulated Annealing (SA) and reports the results

    Parameters
    ----------
    surrogate : Surrogate_wrapper
        tuned surrogate model
    dim: int
        dimension of the problem
    max_iters: int
        maximum number of iterations for SA
    temp: float
        parameter of Simulated Annealing (can be changed or tuned for other problems)
        max temperature. It controls the exploration rate of SA
    alpha:
        parameter of Simulated Annealing (can be changed or tuned for other problems)
        cooling rate. It determines how quickly temp decays

    Returns
    ----------
    x_best: np.ndarray
        optimal solution
    y_best: float
        optimal objective function value
    trace: list[tuple]
        optimization log. Updates at the end of each iteration.
    trace_x: list[np.ndarray]
        tracks how X changes during each iteration.
    """

    # generate random starting point
    x_curr = np.random.randint(0, 3, size=dim)

    y_curr = surrogate.predict(x_curr)
    x_best, y_best = x_curr.copy(), y_curr

    trace = [(0, y_best)]
    trace_x = [x_curr.copy()]

    for i in range(max_iters):
        # Optimization loop

        x_new = x_curr.copy()
        idx = random.randint(0, dim - 1)
        # perturb a random dimension to get new x
        x_new[idx] = (x_new[idx] + random.choice([-1, 1])) % 3
        y_new = surrogate.predict(x_new)

        delta = y_new - y_curr
        if delta < 0 or np.random.rand() < np.exp(-delta / temp):
            x_curr, y_curr = x_new, y_new
            if y_curr < y_best:
                x_best, y_best = x_new.copy(), y_new

        trace.append((i, y_best))
        trace_x.append(x_curr.copy())
        temp *= alpha
    return x_best, y_best, trace, trace_x


def run_optimization(
    X: np.ndarray,
    y: np.ndarray,
    model_type: str = "rbf",
    n_runs: int = 10,
    max_iters: int = 1000,
    bootstrap_size: int = 100,
    out_folder: str = ".",
):
    """
    Bootstraps data, runs optimization and postprocesses the results.

    Parameters
    ----------
    X: np.ndarray
        Raw design configuration
        Dimension N by d, where N is the number of simulations,
                       and d is the number of design variables
    y: np.ndarray
        Raw array of quantity of interest
        Dimension N by 1, where N is the number of simulations

    model_type : str
        'rbf' for RBFInterpolator
        'rf' for Random Forest
        'nn' for Neural Network

    n_runs :int
        number of bootstraps
    max_iters: int
        maximum number of SA iterations
    bootstrap_size : int
        size of bootstrap samples
    out_folder: str
        folder where to output the results

    Returns
    ----------
    """

    all_x = []
    all_y = []
    all_traces = np.zeros((n_runs, max_iters + 1))
    all_trace_x = []
    rng = np.random.default_rng(42)
    # Bootstrap the data
    bootstrap_idxs = [
        rng.choice(len(X), size=bootstrap_size, replace=False)
        for _ in range(n_runs)
    ]

    i = 0
    for idxs in bootstrap_idxs:
        # Tune the model for each bootstrap
        X_sub, y_sub = X[idxs], y[idxs]
        if model_type == "rbf":
            params = tune_rbf(X_sub, y_sub)
        elif model_type == "rf":
            params = tune_rf(X_sub, y_sub)
        elif model_type == "nn":
            encoder = OneHotEncoder(sparse_output=False)
            X_encoded = encoder.fit_transform(X_sub)
            params = tune_nn(X_encoded, y_sub)
        else:
            raise ValueError("Invalid model_type")

        # build the surrogate model
        surrogate = Surrogate_wrapper(model_type, X_sub, y_sub, params)

        # run optimization
        x_best, y_best, trace, trace_x = simulated_annealing_surrogate(
            surrogate, dim=X.shape[1], max_iters=max_iters
        )

        trace_y = [y for _, y in trace]
        all_traces[i, :] = trace_y
        all_x.append(x_best)
        all_y.append(y_best)
        all_trace_x.append(trace_x)
        i = i + 1

    # Compute the best y across all the bootstraps
    best_index = np.argmin(all_y)
    best_x = all_x[best_index]
    best_y = all_y[best_index]
    # Save the best solution
    df = pd.DataFrame(
        [
            {
                **{f"x{i}": best_x[i] for i in range(len(best_x))},
                "best_y": best_y,
            }
        ]
    )
    df.to_csv(
        os.path.join(
            out_folder,
            f"best_bootstrap_solution_{model_type}_size_{bootstrap_size}.csv",
        ),
        index=False,
    )

    # Make the mean-CI plot for the objective function and save it
    mean_trace = np.mean(-1 * all_traces, axis=0)
    std_trace = np.std(all_traces, axis=0)
    lower_bound = mean_trace - 1.96 * std_trace / np.sqrt(n_runs)
    upper_bound = mean_trace + 1.96 * std_trace / np.sqrt(n_runs)
    iterations = np.arange(max_iters + 1)

    plt.figure(figsize=(8, 6))
    plt.plot(
        iterations,
        mean_trace,
        label=f"Mean Convergence {model_type.upper()}",
        color="blue",
    )
    plt.fill_between(
        iterations,
        lower_bound,
        upper_bound,
        color="blue",
        alpha=0.3,
        label="95% CI",
    )
    pretty_labels(
        "Iteration",
        r"Predicted QOI [kg$^2$/kWh$^2$]",
        fontsize=20,
        title=f"Mean Convergence with 95% Confidence Interval ({model_type.upper()})",
        grid=True,
        fontname="Times",
    )
    pretty_legend(fontsize=20, fontname="Times")
    plt.savefig(
        os.path.join(
            out_folder,
            f"Mean_Convergence_plot_{model_type}_size_{bootstrap_size}.png",
        ),
        dpi=300,
    )
    plt.savefig(
        os.path.join(
            out_folder,
            f"Mean_Convergence_plot_{model_type}_size_{bootstrap_size}.pdf",
        ),
    )
    plt.show()

    # Compute the distance of intermediate x from the optimal solution
    optimal_x = np.ones(X.shape[1], dtype=int)
    l1_traces = []
    for t in all_trace_x:
        distances = [np.sum(np.abs(np.array(x) - optimal_x)) for x in t]
        l1_traces.append(distances)

    """ compute and make the mean-CI plot for the distance between the intermdeiate   
    solution and the optimal solution """
    l1_traces = np.array(l1_traces)
    mean_l1 = np.mean(l1_traces, axis=0)
    std_l1 = np.std(l1_traces, axis=0)
    lower = mean_l1 - 1.96 * std_l1 / np.sqrt(len(trace_x))
    upper = mean_l1 + 1.96 * std_l1 / np.sqrt(len(trace_x))
    iterations = np.arange(len(mean_l1))
    plt.figure(figsize=(8, 4))
    plt.plot(
        iterations,
        mean_l1,
        label="Mean L1 Distance from Optimal",
        color="darkred",
    )
    plt.fill_between(
        iterations, lower, upper, alpha=0.3, color="darkred", label="95% CI"
    )
    pretty_labels(
        "Iteration",
        "L1 Distance from Optimal",
        fontsize=16,
        title=f"Convergence Toward Optimal Solution (L1 Norm)",
        grid=True,
        fontname="Times",
    )
    pretty_legend(fontsize=16, fontname="Times")
    plt.tight_layout()
    plt.savefig(
        os.path.join(
            out_folder,
            f"Mean_L1_distance_{model_type}_size_{bootstrap_size}.png",
        ),
        dpi=300,
    )
    plt.savefig(
        os.path.join(
            out_folder,
            f"Mean_L1_distance_{model_type}_size_{bootstrap_size}.pdf",
        )
    )
    plt.show()


if __name__ == "__main__":

    # studies = ["study_scaleup_0_4vvm_3000W", "study_scaleup_0_1vvm_6000W", "study_0_4vvm_1W"]
    # studies = ["study_scaleup_0_1vvm_6000W", "study_0_4vvm_1W"]
    studies = ["study_scaleup_0_4vvm_3000W", "study_0_4vvm_1W"]

    for study in studies:
        # Read data from the csv file.
        X_raw_data = pd.read_csv(os.path.join(study, "Xdata.csv"))
        y_raw_data = pd.read_csv(os.path.join(study, "ydata.csv"))

        X = X_raw_data.values
        y = y_raw_data.iloc[:, :-1].values
        # We want to maximize, so we minimize the opposite
        y = y * -1

        # The function will build, tune, and optimize the surrogate model and postprocess the results.
        run_optimization(
            X,
            y,
            model_type="rbf",
            n_runs=10,
            max_iters=1000,
            bootstrap_size=150,
            out_folder=study,
        )
        # run_optimization(
        #    X,
        #    y,
        #    model_type="rf",
        #    n_runs=10,
        #    max_iters=1000,
        #    bootstrap_size=150,
        #    out_folder=study,
        # )
        # run_optimization(
        #    X, y, model_type="nn", n_runs=10, max_iters=1000, bootstrap_size=150, out_folder=study
        # )



import os

import numpy as np

from bird.utilities.ofio import *


def writeGvars(inletA, liqVol):
    filename_tmp = os.path.join("constant", "globalVars_temp")
    with open(filename_tmp, "r+") as f:
        lines = f.readlines()
    filename = os.path.join("constant", "globalVars")
    with open(filename, "w+") as f:
        for line in lines:
            if line.startswith("inletA"):
                f.write(f"inletA\t{inletA:g};\n")
            elif line.startswith("liqVol"):
                f.write(f"liqVol\t{liqVol:g};\n")
            else:
                f.write(line)


def readInletArea():
    filename = os.path.join(
        "postProcessing",
        "patchIntegrate(patch=inlet,field=alpha.gas)",
        "0",
        "surfaceFieldValue.dat",
    )
    with open(filename, "r+") as f:
        lines = f.readlines()
    return float(lines[4].split()[-1])


def getLiqVol():
    cellCentres = readMesh(os.path.join(".", f"meshCellCentres_0.obj"))
    volume_field = readOFScal(os.path.join("0", "V"), len(cellCentres))[
        "field"
    ]
    alpha_field = readOFScal(
        os.path.join("0", "alpha.liquid"), len(cellCentres)
    )["field"]
    return np.sum(volume_field * alpha_field)


if __name__ == "__main__":
    A = readInletArea()
    V = getLiqVol()
    writeGvars(A, V)



import logging

import numpy as np

logger = logging.getLogger(__name__)


class Mixer:
    def __init__(self):
        self.x = None
        self.y = None
        self.z = None
        self.rad = 0.015
        self.power = 300
        self.start_time = 1.0
        self.smear = 3
        self.sign = None
        self.normal_dir = None
        self.ready = False

    def update_from_expl_dict(self, mixer_dict):
        if "x" in mixer_dict:
            self.x = mixer_dict["x"]
        if "y" in mixer_dict:
            self.y = mixer_dict["y"]
        if "z" in mixer_dict:
            self.z = mixer_dict["z"]
        if "rad" in mixer_dict:
            self.rad = mixer_dict["rad"]
        if "power" in mixer_dict:
            self.power = mixer_dict["power"]
        if "sign" in mixer_dict:
            self.sign = mixer_dict["sign"]
        if "smear" in mixer_dict:
            self.smear = mixer_dict["smear"]
        if "start_time" in mixer_dict:
            self.start_time = mixer_dict["start_time"]
        if "normal_dir" in mixer_dict:
            self.normal_dir = mixer_dict["normal_dir"]
        self.check_status()

    def update_from_loop_dict(self, mixer_dict, geom_dict, mesh_dict=None):
        segment = geom_dict["segments"][mixer_dict["branch_id"]]
        pos = segment["start"] + mixer_dict["frac_space"] * segment["conn"]
        self.x = pos[0]
        self.y = pos[1]
        self.z = pos[2]
        self.normal_dir = segment["normal_dir"]
        if "rad" in mixer_dict:
            self.rad = min(mixer_dict["rad"], segment["max_rad"])
        else:
            self.rad = segment["max_rad"] * 0.7
        if "power" in mixer_dict:
            self.power = mixer_dict["power"]
        if "sign" in mixer_dict:
            self.sign = mixer_dict["sign"]
        if "start_time" in mixer_dict:
            self.start_time = mixer_dict["start_time"]
        if "normal_dir" in mixer_dict:
            self.normal_dir = mixer_dict["normal_dir"]
        if mesh_dict is not None:
            if self.normal_dir == 0:
                min_mesh_transv = min(
                    mesh_dict["Blockwise"]["y"], mesh_dict["Blockwise"]["z"]
                )
            elif self.normal_dir == 1:
                min_mesh_transv = min(
                    mesh_dict["Blockwise"]["x"], mesh_dict["Blockwise"]["z"]
                )
            elif self.normal_dir == 2:
                min_mesh_transv = min(
                    mesh_dict["Blockwise"]["x"], mesh_dict["Blockwise"]["y"]
                )
            self.smear = min_mesh_transv // 3
        self.check_status(blocks=segment["blocks"])

    def check_status(self, blocks=None):
        if (
            self.x is None
            or self.y is None
            or self.z is None
            or self.normal_dir is None
            or ((not self.sign == "+") and (not self.sign == "-"))
        ):
            self.ready = False
        else:
            logger.info(
                f"\n\tpos({self.x:.2g}, {self.y:.2g}, {self.z:.2g})"
                + f"\n\tnormal_dir {self.normal_dir}"
                + f"\n\trad {self.rad:.2g}"
                + f"\n\tpower {self.power:.2g}"
                + f"\n\tsign {self.sign}"
                + f"\n\tsmear {self.smear}"
                + f"\n\tstart_time {self.start_time:.2g}"
            )
            if blocks is not None:
                logger.info(f"\tbranch = {blocks}")

            self.ready = True



import os


def write_preamble(output_folder):
    with open(os.path.join(output_folder, "fvModels"), "w+") as f:
        f.write("FoamFile\n")
        f.write("{\n")
        f.write("\tversion  9.0;\n")
        f.write("\tformat   ascii;\n")
        f.write("\tclass    dictionary;\n")
        f.write('\tlocation "constant";\n')
        f.write("\tobject   fvModels;\n")
        f.write("}\n\n")
        f.write("codedSource\n")
        f.write("{\n")
        f.write("\ttype\tcoded;\n")
        f.write("\tselectionMode\tall;\n")
        f.write("\tfield\tU.liquid;\n")
        f.write("\tname\tsourceTime;\n\n")
        f.write("\tcodeInclude\n")
        f.write("\t#{\n")
        f.write('\t\t#include "dynamicMix_util.H"\n')
        f.write("\t#};\n")
        f.write("\n")
        f.write("\tcodeOptions\n")
        f.write("\t#{\n")
        f.write("\t\t-I${FOAM_CASE}/constant\n")
        f.write("\t#};\n")

        f.write("\tcodeAddAlphaRhoSup\n")
        f.write("\t#{\n")
        f.write("\t\tconst Time& time = mesh().time();\n")
        f.write("\t\tconst scalarField& V = mesh().V();\n")
        f.write("\t\tvectorField& Usource = eqn.source();\n")
        f.write("\t\tconst vectorField& C = mesh().C();\n")
        f.write("\t\tconst volScalarField& rhoL =\n")
        f.write(
            '\t\t\tmesh().lookupObject<volScalarField>("thermo:rho.liquid");\n'
        )
        f.write("\t\tconst volScalarField& alphaL =\n")
        f.write('\t\t\tmesh().lookupObject<volScalarField>("alpha.liquid");\n')
        f.write("\t\tconst volVectorField& UL =\n")
        f.write('\t\t\tmesh().lookupObject<volVectorField>("U.liquid");\n')
        f.write("\t\tdouble pi=3.141592654;\n")
        f.write(f"\t\tdouble source_pt_x;\n")
        f.write(f"\t\tdouble source_pt_y;\n")
        f.write(f"\t\tdouble source_pt_z;\n")
        f.write(f"\t\tdouble disk_rad;\n")
        f.write("\t\tdouble disk_area;\n")
        f.write(f"\t\tdouble power;\n")
        f.write(f"\t\tdouble smear_factor;\n")
        f.write(f"\t\tdouble startTime;\n")


def write_mixer(mixer, output_folder):
    with open(os.path.join(output_folder, "fvModels"), "a+") as f:
        f.write(f"\t\tsource_pt_x={mixer.x};\n")
        f.write(f"\t\tsource_pt_y={mixer.y};\n")
        f.write(f"\t\tsource_pt_z={mixer.z};\n")
        f.write(f"\t\tdisk_rad={mixer.rad};\n")
        f.write("\t\tdisk_area=pi*disk_rad*disk_rad;\n")
        f.write(f"\t\tpower={mixer.power};\n")
        f.write(f"\t\tsmear_factor={float(mixer.smear)};\n")
        f.write(f"\t\tstartTime = {mixer.start_time};\n")
        f.write("\t\tif (time.value() > startTime)\n")
        f.write("\t\t{\n")
        f.write("\t\t\t// Get V1\n")
        f.write("\t\t\tdouble source_sign_factor = 1.0;\n")
        f.write("\t\t\tdouble V1 = 0;\n")
        f.write("\t\t\tdouble V2 = 0;\n")
        f.write("\t\t\tdouble rhoV;\n")
        f.write("\t\t\tdouble dist_tol = disk_rad*3;\n")
        f.write("\n")
        f.write("\t\t\tdouble dist_n;\n")
        f.write("\t\t\tdouble upV = 0;\n")
        f.write("\t\t\tdouble uprhoV = 0;\n")
        f.write("\t\t\tdouble upVvol = 0;\n")
        f.write("\t\t\tdouble downV = 0;\n")
        f.write("\t\t\tdouble downrhoV = 0;\n")
        f.write("\t\t\tdouble downVvol = 0;\n")
        f.write("\t\t\tdouble dist2;\n")

        f.write("\t\t\tforAll(C,i)\n")
        f.write("\t\t\t{\n")
        f.write(
            "\t\t\t\tdist2 = (C[i].x()-source_pt_x)*(C[i].x()-source_pt_x);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].y()-source_pt_y)*(C[i].y()-source_pt_y);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].z()-source_pt_z)*(C[i].z()-source_pt_z);\n"
        )
        f.write("\n")
        if mixer.normal_dir == 0:
            f.write("\t\t\t\tdist_n = (C[i].x()-source_pt_x);\n")
        elif mixer.normal_dir == 1:
            f.write("\t\t\t\tdist_n = (C[i].y()-source_pt_y);\n")
        elif mixer.normal_dir == 2:
            f.write("\t\t\t\tdist_n = (C[i].z()-source_pt_z);\n")
        f.write("\n")

        f.write(
            "\t\t\t\tif (dist2 < dist_tol*dist_tol && dist_n < -dist_tol/2) {\n"
        )
        f.write("\t\t\t\t\tupVvol += V[i] * alphaL[i];\n")
        f.write(
            f"\t\t\t\t\tupV += V[i] * alphaL[i] * UL[i][{int(mixer.normal_dir)}];\n"
        )
        f.write("\t\t\t\t\tuprhoV += V[i] * alphaL[i] * rhoL[i];\n")
        f.write("\t\t\t\t}\n")
        f.write(
            "\t\t\t\tif (dist2 < dist_tol*dist_tol && dist_n > dist_tol/2) {\n"
        )
        f.write("\t\t\t\t\tdownVvol += V[i] * alphaL[i];\n")
        f.write(
            f"\t\t\t\t\tdownV += V[i] * alphaL[i] * UL[i][{int(mixer.normal_dir)}];\n"
        )
        f.write("\t\t\t\t\tdownrhoV += V[i] * alphaL[i] * rhoL[i];\n")
        f.write("\t\t\t\t}\n")
        f.write("\t\t\t}\n")
        f.write("\n")
        f.write("\t\t\treduce(uprhoV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(downrhoV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(upV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(downV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(downVvol, sumOp<scalar>());\n")
        f.write("\t\t\treduce(upVvol, sumOp<scalar>());\n")
        f.write("\n")
        f.write("\t\t\tdownV /= downVvol;\n")
        f.write("\t\t\tupV /= upVvol;\n")
        f.write("\t\t\tdownrhoV /= downVvol;\n")
        f.write("\t\t\tuprhoV /= upVvol;\n")
        f.write("\n")
        f.write("\t\t\tif (upV <= 0 && downV <= 0) {\n")
        f.write("\t\t\t\tsource_sign_factor = -1.0;\n")
        f.write("\t\t\t\tV1 = std::abs(upV);\n")
        f.write("\t\t\t\trhoV = uprhoV;\n")
        f.write("\t\t\t} else if (upV >= 0 && downV >= 0) {\n")
        f.write("\t\t\t\tsource_sign_factor = 1.0;\n")
        f.write("\t\t\t\tV1 = std::abs(downV);\n")
        f.write("\t\t\t\trhoV = downrhoV;\n")
        f.write("\t\t\t} else {\n")
        f.write("\t\t\t\tV1 = 0.0;\n")
        if mixer.sign == "+":
            f.write("\t\t\t\tsource_sign_factor = -1.0;\n")
            f.write("\t\t\t\trhoV = uprhoV;\n")
        elif mixer.sign == "-":
            f.write("\t\t\t\tsource_sign_factor = 1.0;\n")
            f.write("\t\t\t\trhoV = downrhoV;\n")
        else:
            error_message = (
                f"mixer.sign = {mixer.sign} but should be '+' or '-'"
            )
            logger.error
        f.write(
            '\t\t\t\tFoam::Info << "[BIRD:DYNMIX WARN] " << "upV = " << upV << " downV = " << downV << " for source at " << source_pt_x << ", " << source_pt_y << ", " << source_pt_z <<  endl;\n'
        )
        f.write("\t\t\t}\n")
        f.write(
            '\t\t\tFoam::Info << "[BIRD:DYNMIX INFO] V1 = " << V1 << endl;\n'
        )
        f.write("\t\t\t\n")
        f.write("\t\t\t// Get V2\n")
        f.write("\t\t\tV2 = findV2(power, rhoV, disk_area, V1);\n")
        f.write("\n")
        f.write("\t\t\tforAll(C,i)\n")
        f.write("\t\t\t{\n")
        f.write(
            "\t\t\t\tdouble Thrust=0.5*rhoL[i]*(V2*V2 - V1*V1)*disk_area;\n"
        )
        f.write(
            "\t\t\t\tdouble dist2=(C[i].x()-source_pt_x)*(C[i].x()-source_pt_x);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].y()-source_pt_y)*(C[i].y()-source_pt_y);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].z()-source_pt_z)*(C[i].z()-source_pt_z);\n"
        )

        f.write("\t\t\t\tdouble epsilon=pow(V[i],0.33333)*smear_factor;\n")
        f.write(
            "\t\t\t\tdouble sourceterm=alphaL[i]*(Thrust/pow(pi,1.5)/pow(epsilon,3.0))*\n"
        )
        f.write("\t\t\t\t\texp(-dist2/(epsilon*epsilon));\n")

        f.write(
            f"\t\t\t\tUsource[i][{int(mixer.normal_dir)}] -=  source_sign_factor*sourceterm*V[i];\n"
        )

        f.write("\t\t\t}\n")
        f.write("\t\t}\n")


def write_mixer_force_sign(mixer, output_folder):
    with open(os.path.join(output_folder, "fvModels"), "a+") as f:
        f.write(f"\t\tsource_pt_x={mixer.x};\n")
        f.write(f"\t\tsource_pt_y={mixer.y};\n")
        f.write(f"\t\tsource_pt_z={mixer.z};\n")
        f.write(f"\t\tdisk_rad={mixer.rad};\n")
        f.write("\t\tdisk_area=pi*disk_rad*disk_rad;\n")
        f.write(f"\t\tpower={mixer.power};\n")
        f.write(f"\t\tsmear_factor={float(mixer.smear)};\n")
        f.write(f"\t\tstartTime = {mixer.start_time};\n")
        f.write("\t\tif (time.value() > startTime)\n")
        f.write("\t\t{\n")
        f.write("\t\t\t// Get V1\n")
        f.write("\t\t\tdouble source_sign_factor = 1.0;\n")
        f.write("\t\t\tdouble V1 = 0;\n")
        f.write("\t\t\tdouble V2 = 0;\n")
        f.write("\t\t\tdouble rhoV;\n")
        f.write("\t\t\tdouble dist_tol = disk_rad*3;\n")
        f.write("\n")
        f.write("\t\t\tdouble dist_n;\n")
        f.write("\t\t\tdouble upV = 0;\n")
        f.write("\t\t\tdouble uprhoV = 0;\n")
        f.write("\t\t\tdouble upVvol = 0;\n")
        f.write("\t\t\tdouble downV = 0;\n")
        f.write("\t\t\tdouble downrhoV = 0;\n")
        f.write("\t\t\tdouble downVvol = 0;\n")
        f.write("\t\t\tdouble dist2;\n")

        f.write("\t\t\tforAll(C,i)\n")
        f.write("\t\t\t{\n")
        f.write(
            "\t\t\t\tdist2 = (C[i].x()-source_pt_x)*(C[i].x()-source_pt_x);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].y()-source_pt_y)*(C[i].y()-source_pt_y);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].z()-source_pt_z)*(C[i].z()-source_pt_z);\n"
        )
        f.write("\n")
        if mixer.normal_dir == 0:
            f.write("\t\t\t\tdist_n = (C[i].x()-source_pt_x);\n")
        elif mixer.normal_dir == 1:
            f.write("\t\t\t\tdist_n = (C[i].y()-source_pt_y);\n")
        elif mixer.normal_dir == 2:
            f.write("\t\t\t\tdist_n = (C[i].z()-source_pt_z);\n")
        f.write("\n")

        f.write(
            "\t\t\t\tif (dist2 < dist_tol*dist_tol && dist_n < -dist_tol/2) {\n"
        )
        f.write("\t\t\t\t\tupVvol += V[i] * alphaL[i];\n")
        f.write(
            f"\t\t\t\t\tupV += V[i] * alphaL[i] * UL[i][{int(mixer.normal_dir)}];\n"
        )
        f.write("\t\t\t\t\tuprhoV += V[i] * alphaL[i] * rhoL[i];\n")
        f.write("\t\t\t\t}\n")
        f.write(
            "\t\t\t\tif (dist2 < dist_tol*dist_tol && dist_n > dist_tol/2) {\n"
        )
        f.write("\t\t\t\t\tdownVvol += V[i] * alphaL[i];\n")
        f.write(
            f"\t\t\t\t\tdownV += V[i] * alphaL[i] * UL[i][{int(mixer.normal_dir)}];\n"
        )
        f.write("\t\t\t\t\tdownrhoV += V[i] * alphaL[i] * rhoL[i];\n")
        f.write("\t\t\t\t}\n")
        f.write("\t\t\t}\n")
        f.write("\n")
        f.write("\t\t\treduce(uprhoV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(downrhoV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(upV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(downV, sumOp<scalar>());\n")
        f.write("\t\t\treduce(downVvol, sumOp<scalar>());\n")
        f.write("\t\t\treduce(upVvol, sumOp<scalar>());\n")
        f.write("\n")
        f.write("\t\t\tdownV /= downVvol;\n")
        f.write("\t\t\tupV /= upVvol;\n")
        f.write("\t\t\tdownrhoV /= downVvol;\n")
        f.write("\t\t\tuprhoV /= upVvol;\n")
        f.write("\n")
        if mixer.sign == "+":
            f.write("\t\t\tsource_sign_factor = -1.0;\n")
            f.write("\t\t\tif (upV >= 0){\n")
            f.write("\t\t\t\tV1 = 0.0;\n")
            f.write("\t\t\t} else {\n")
            f.write("\t\t\t\tV1 = std::abs(upV);\n")
            f.write("\t\t\t}\n")
            f.write("\t\t\trhoV = uprhoV;\n")
        elif mixer.sign == "-":
            f.write("\t\t\tsource_sign_factor = 1.0;\n")
            f.write("\t\t\tif (downV <= 0){\n")
            f.write("\t\t\t\tV1 = 0.0;\n")
            f.write("\t\t\t} else {\n")
            f.write("\t\t\t\tV1 = std::abs(downV);\n")
            f.write("\t\t\t}\n")
            f.write("\t\t\trhoV = downrhoV;\n")
        # f.write("\t\t\t}\n")
        f.write(
            '\t\t\tFoam::Info << "[BIRD:DYNMIX INFO] V1 = " << V1 << endl;\n'
        )
        f.write("\t\t\t\n")
        f.write("\t\t\t// Get V2\n")
        f.write("\t\t\tV2 = findV2(power, rhoV, disk_area, V1);\n")
        f.write("\n")
        f.write("\t\t\tforAll(C,i)\n")
        f.write("\t\t\t{\n")
        f.write(
            "\t\t\t\tdouble Thrust=0.5*rhoL[i]*(V2*V2 - V1*V1)*disk_area;\n"
        )
        f.write(
            "\t\t\t\tdouble dist2=(C[i].x()-source_pt_x)*(C[i].x()-source_pt_x);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].y()-source_pt_y)*(C[i].y()-source_pt_y);\n"
        )
        f.write(
            "\t\t\t\tdist2 += (C[i].z()-source_pt_z)*(C[i].z()-source_pt_z);\n"
        )

        f.write("\t\t\t\tdouble epsilon=pow(V[i],0.33333)*smear_factor;\n")
        f.write(
            "\t\t\t\tdouble sourceterm=alphaL[i]*(Thrust/pow(pi,1.5)/pow(epsilon,3.0))*\n"
        )
        f.write("\t\t\t\t\texp(-dist2/(epsilon*epsilon));\n")

        f.write(
            f"\t\t\t\tUsource[i][{int(mixer.normal_dir)}] -=  source_sign_factor*sourceterm*V[i];\n"
        )

        f.write("\t\t\t}\n")
        f.write("\t\t}\n")


def write_end(output_folder):
    with open(os.path.join(output_folder, "fvModels"), "a+") as f:
        f.write("\t#};\n")
        f.write("};\n")



import logging
import os
import re

logger = logging.getLogger(__name__)


def makeRecursiveFolder(path):
    folder_list = path.split("/")
    localFolder = ""
    for folder in folder_list:
        localFolder = os.path.join(localFolder, folder)
        os.makedirs(localFolder, exist_ok=True)


def getManyFolders(rootFolder, prefix="flat_donut"):
    # Read Time
    fold_tmp = os.listdir(rootFolder)
    fold_num = []
    # remove non floats
    for i, entry in reversed(list(enumerate(fold_tmp))):
        if not entry.startswith(prefix) or entry.endswith("N2"):
            a = fold_tmp.pop(i)
            # print('removed ', a)
    for entry in fold_tmp:
        num = re.findall(r"\d+", entry)
        if len(num) > 1:
            msg = f"Cannot find num of folder {entry}."
            msg += "\nDo not trust the spearman stat"
            logger.warning(msg)
        else:
            fold_num.append(int(num[0]))

    sortedFold = [
        x for _, x in sorted(zip(fold_num, fold_tmp), key=lambda pair: pair[0])
    ]
    return sortedFold



import os
import shutil
import sys

import numpy as np

sys.path.append("util")
import argument_inhomo
from fromMomtoPdf import *


def writeFfield(filename, fieldname, xcent, zcent, schedule, defaultVal, ind):
    fw = open(filename, "w+")
    # Write Header
    fw.write("FoamFile\n")
    fw.write("{\n")
    fw.write("    format      ascii;\n")
    fw.write("    class       volScalarField;\n")
    fw.write(f"    object      {fieldname};\n")
    fw.write("}\n")
    fw.write("\n")
    fw.write("dimensions      [0 0 0 0 0 0 0];\n")
    fw.write("\n")
    fw.write("internalField   uniform 0;\n")
    fw.write("\n")
    fw.write("boundaryField\n")
    fw.write("{\n")
    # All the walls first
    zgBound = ["walls", "outlet"]
    for bound in zgBound:
        fw.write(f"    {bound}\n")
        fw.write("    {" + "\n")
        fw.write("        type            zeroGradient;\n")
        fw.write("    }" + "\n")
    cmBound = ["inlet"]
    for bound in cmBound:
        fw.write(f"    {bound}\n")
        fw.write("    {" + "\n")
        fw.write("        type             codedMixed;\n")
        fw.write(f"        refValue         uniform {defaultVal};\n")
        fw.write("        refGradient      uniform 0;\n")
        fw.write("        valueFraction    uniform 1.0;\n")
        fw.write(f"        redirectType     {fieldname}{bound};\n")
        fw.write(f"        name             {fieldname}{bound};\n")
        fw.write("\n")
        fw.write("        code\n")
        fw.write("        #{\n")
        fw.write("            const fvPatch& boundaryPatch = patch();" + "\n")
        fw.write(
            "            const vectorField& Cf = boundaryPatch.Cf();" + "\n"
        )
        fw.write("            forAll(Cf, faceI)\n")
        fw.write("            {\n")
        fw.write(
            f"                scalar xpos = boundaryPatch.Cf()[faceI][0]-{xcent};\n"
        )
        fw.write(
            f"                scalar zpos = boundaryPatch.Cf()[faceI][2]-{zcent};\n"
        )
        fw.write(
            "                scalar rad = std::sqrt(xpos*xpos + zpos*zpos);\n"
        )
        fw.write(f'                if( rad < {schedule["r_const"]} )\n')
        fw.write("                {\n")
        fw.write(
            f'                    this->refValue()[faceI] = {schedule["f_in"][ind]};\n'
        )
        fw.write("                    this->refGrad()[faceI] = 0.0;\n")
        fw.write("                    this->valueFraction()[faceI] = 1.0;\n")
        fw.write(f'                }} else if (rad < {schedule["r_end"]}){{\n')
        fw.write(
            f'                    this->refValue()[faceI] = {schedule["f_in"][ind]} + (rad - {schedule["r_const"]})/({schedule["r_end"] - schedule["r_const"]}) * ({schedule["f_out"][ind] - schedule["f_in"][ind]});\n'
        )
        fw.write("                    this->refGrad()[faceI] = 0.0;\n")
        fw.write("                    this->valueFraction()[faceI] = 1.0;\n")
        fw.write("                } else {\n")
        fw.write(
            f'                    this->refValue()[faceI] = {schedule["f_out"][ind]};\n'
        )
        fw.write("                    this->refGrad()[faceI] = 0.0;\n")
        fw.write("                    this->valueFraction()[faceI] = 1.0;\n")
        fw.write("                }\n")
        fw.write("            }\n")
        fw.write("        #};\n")
        fw.write("        codeInclude\n")
        fw.write("        #{\n")
        fw.write("        #};\n")
        fw.write("\n")
        fw.write("        codeOptions" + "\n")
        fw.write("        #{" + "\n")
        fw.write("        #};" + "\n")
        fw.write("    }" + "\n")

        fw.write("}\n")
        fw.close()


def binInfo(phasePropFile):
    f = open(phasePropFile, "r")
    nFields = 0
    diam = []
    value = []
    fname = []
    lines = f.readlines()
    for iline, line in enumerate(lines):
        if "sizeGroups" in line.strip():
            startLine = iline
            break
    for line in lines[iline + 1 :]:
        if line.strip().startswith("f"):
            nFields += 1
            fname.append(line.strip().split()[0])
            indSem = line.strip().find(";")
            indSem2 = line.strip()[indSem + 1 :].find(";")
            indD = line.strip().find("dSph", 0, indSem)
            try:
                diamStr = line.strip()[indD:indSem].split()[1]
                diam.append(float(diamStr))
            except ValueError:
                print(f"ERROR: Could not convert {diamStr} to float")
            try:
                valStr = line.strip()[indSem + 1 :][:indSem2].split()[1]
                value.append(float(valStr))
            except ValueError:
                print(f"ERROR: Could not convert {valStr} to float")
        elif line.strip().startswith("(") or line.strip().startswith(")"):
            pass
        elif line.strip().startswith("}"):
            break
    f.close()
    bin_size = np.diff(np.array(diam))
    ave_bin_size = np.mean(bin_size)
    std_bin_size = np.std(bin_size)
    if std_bin_size > (ave_bin_size * 1e-6):
        print(f"ERROR: Do not use with non uniform bin spacing")
    return nFields, diam, ave_bin_size, fname, value


if __name__ == "__main__":
    args = argument_inhomo.initArgs()
    print("ASSUMPTIONS:")
    print("\t-constant bin size")
    print("\t-y axis is the gravity axis")
    print("\t-diameters are ordered\n")

    schedule = {
        "r_const": args.r_const,
        "r_end": args.r_end,
        "pore_in": args.pore_in,
        "pore_out": args.pore_out,
    }

    phasePropFile = "constant/phaseProperties"
    nf, diam, bin_size, fname, value = binInfo(phasePropFile)

    import matplotlib.pyplot as plt

    fig = plt.figure()
    for i in range(20):
        meanTar, stdTar = poreDiamCorr(
            dp=schedule["pore_out"], ds=0.15, Ugs=0.01
        )
        f_in = get_f_vals(meanTar, stdTar, np.array(diam), verb=args.verbose)
        plt.plot(diam, f_in)
    plt.show()



import logging
import os
import re

import numpy as np

logger = logging.getLogger(__name__)


def readMesh(filename: str) -> np.ndarray:
    """
    Reads cell center location from meshCellCentres_X.obj

    Parameters
    ----------
    filename: str
        meshCellCentres_X.obj filename

    returns
    ----------
    cell_centers: np.ndarray
        Array (N,3) representing the cell centers (N is number of cells)

    """
    assert "meshCellCentres" in filename
    assert ".obj" in filename
    cell_centers = np.loadtxt(filename, usecols=(1, 2, 3))
    return cell_centers


def ofvec2arr(vec: str) -> np.ndarray:
    """
    Converts a vector written as a string into a numpy array

    Parameters
    ----------
    vec: str
        Vector written as string
        Must start with "("
        Must end with ")"

    returns
    ----------
    vec_array: np.ndarray
        Array (3,) representing the vector


    """
    vec = vec.strip()
    assert vec[0] == "("
    assert vec[-1] == ")"

    vec_list = vec[1:-1].split()
    vec_array = np.array([float(entry) for entry in vec_list])
    return vec_array


def is_comment(line: str) -> bool:
    """
    Checks if line is a comment

    Parameters
    ----------
    line: str
        Line of file

    returns
    ----------
    is_comment: bool
        True if line is a comment
        False if line is not a comment
    """
    is_comment = False

    sline = line.strip()
    if sline.startswith("//"):
        is_comment = True
    elif sline.startswith("/*"):
        is_comment = True
    return is_comment


def read_meta_data(filename: str, mode: str | None = None) -> dict:
    """
    Read meta data from field outputted by OpenFOAM in ASCII format

    Parameters
    ----------
    filename: str
        Field filename
    mode: str | None
        If "scalar", expects a scalar field
        If "vector", expects a vector field
        If None, obtained from field header

    returns
    ----------
    meta_data: dict
        Dictionary that contain info about the scalar field
            ============= =====================================================
            Key           Description (type)
            ============= =====================================================
            name          Name of the field (*str*)
            n_cells       Number of computational cells (*int*)
            uniform       Whether the field is uniform (*bool*)
            uniform_value Uniform value if uniform field (*float* | *np.ndarray*)
            type          "vector" or "scalar" (*str*)
            ============= =====================================================
    """
    meta_data = {}
    meta_data["type"] = mode

    # Read meta data
    with open(filename, "r") as f:
        header_done = False
        iline = 0
        while not header_done:
            line = f.readline()
            if not is_comment(line):
                # Get field type
                if (line.strip().startswith("class")) and (";" in line):
                    sline = line.strip().split()
                    field_type = sline[1][:-1]
                    if field_type == "volVectorField":
                        field_type = "vector"
                    elif field_type == "volScalarField":
                        field_type = "scalar"
                    else:
                        raise NotImplementedError
                    if mode is not None:
                        assert field_type == mode
                    meta_data["type"] = field_type
                # Get field name
                if (line.strip().startswith("object")) and (";" in line):
                    sline = line.strip().split()
                    field_name = sline[1][:-1]
                    meta_data["name"] = field_name

                # Check if uniform
                if line.strip().startswith("internalField"):
                    if "nonuniform" in line:
                        meta_data["uniform"] = False
                        iline += 1
                        # read until no comments
                        comment = True
                        while comment:
                            count_line = f.readline().strip()
                            if not is_comment(count_line):
                                comment = False
                        try:
                            n_cells = int(count_line)
                            meta_data["n_cells"] = n_cells
                            header_done = True
                        except ValueError:
                            raise ValueError(
                                f"Expected integer number of cells on line {iline}, got: '{count_line}'"
                            )
                    elif "uniform" in line:
                        meta_data["uniform"] = True
                        if meta_data["type"] == "scalar":
                            sline = line.split()
                            unif_value = float(sline[-1].strip(";"))
                        elif meta_data["type"] == "vector":
                            sline = line.split()
                            for ientry, entry in enumerate(sline):
                                if ";" in entry:
                                    ind_end = ientry
                                    break
                            line_cropped = " ".join(sline[2 : ientry + 1])
                            unif_value = ofvec2arr(line_cropped.strip(";"))
                        else:
                            raise NotImplementedError(
                                f"Mode {mode} is unknown"
                            )
                        meta_data["uniform_value"] = unif_value
                        header_done = True

            if len(line) == 0 and (not header_done):
                raise ValueError(
                    f"File {filename} ends before meta-data found"
                )

    return meta_data


def readOFScal(
    filename: str,
    n_cells: int | None = None,
    n_header: int | None = None,
    meta_data: dict | None = None,
) -> dict:
    """
    Read a scalar field outputted by OpenFOAM in ASCII format

    Parameters
    ----------
    filename: str
        Field filename
    n_cells : int | None
        Number of computational cells in the domain
    n_header : int | None
        Number of header lines
    meta_data : dict | None
        meta data dictionary
        If None, it is read from filename

    returns
    ----------
    data: dict
        Dictionary that contain info about the scalar field
            ======== =====================================================
            Key      Description (type)
            ======== =====================================================
            field    For nonuniform fields, array of size the number of cells (*np.ndarray*).
                     For uniform fields with a specified n_cells,
                          array of size the number of cells (*np.ndarray*).
                     For uniform fields, a scalar value (*float*)
            name     Name of the scalar field (*str*)
            n_cells  Number of computational cells (*int*)
            n_header Number of header lines (*int*)
            ======== =====================================================
    """
    field = None

    if meta_data is None:
        # Read meta data
        meta_data = read_meta_data(filename, mode="scalar")

    # Set field
    if meta_data["uniform"]:
        if n_cells is None:
            field = meta_data["uniform_value"]
        else:
            field = meta_data["uniform_value"] * np.ones(n_cells)
    else:
        n_cells = meta_data["n_cells"]

        # Get header size
        if n_header is None:
            oldline = ""
            newline = ""
            iline = 0
            eof = False
            with open(filename, "r") as f:
                while iline < 100 and (not eof):
                    line = f.readline()
                    if len(line) == 0:
                        eof = True
                    oldline = newline
                    newline = line
                    if str(n_cells) in oldline and "(" in newline:
                        n_header = iline + 1
                        break
                    iline += 1
                else:
                    raise ValueError(
                        "Could not find a sequence {n_cells} and '(' in file ({filename})"
                    )

        # Rapid field read
        try:
            field = np.loadtxt(filename, skiprows=n_header, max_rows=n_cells)
        except Exception as err:
            error_msg = f"Issue when reading {filename}"
            logger.error(error_msg)
            raise err(error_msg)

    return {
        "field": field,
        "name": meta_data["name"],
        "n_cells": n_cells,
        "n_header": n_header,
    }


def readOFVec(
    filename: str,
    n_cells: int | None = None,
    n_header: int | None = None,
    meta_data: dict | None = None,
) -> dict:
    """
    Read a vector field outputted by OpenFOAM in ASCII format

    Parameters
    ----------
    filename: str
        Vector field filename
    n_cells : int | None
        Number of computational cells in the domain
    n_header : int | None
        Number of header lines
    meta_data : dict | None
        meta data dictionary
        If None, it is read from filename

    returns
    ----------
    data: dict
        Dictionary that contain info about the scalar field
            ======== =====================================================
            Key      Description (type)
            ======== =====================================================
            field    For nonuniform fields, array of size the number of cells by 3 (*np.ndarray*).
                     For uniform fields with a specified n_cells,
                          array of size the number of cells by 3 (*np.ndarray*).
                     For uniform fields, a scalar value (*float*)
            name     Name of the field (*str*)
            n_cells  Number of computational cells (*int*)
            n_header Number of header lines (*int*)
            ======== =====================================================
    """
    field = None

    if meta_data is None:
        # Read meta data
        meta_data = read_meta_data(filename, mode="vector")

    # Set field
    if meta_data["uniform"]:
        if n_cells is None:
            field = meta_data["uniform_value"]
        else:
            field = meta_data["uniform_value"] * np.ones((n_cells, 3))
    else:
        n_cells = meta_data["n_cells"]
        if n_header is None:
            oldline = ""
            newline = ""
            iline = 0
            eof = False
            with open(filename, "r") as f:
                while iline < 100 and (not eof):
                    line = f.readline()
                    if len(line) == 0:
                        eof = True
                    oldline = newline
                    newline = line
                    if str(n_cells) in oldline and "(" in newline:
                        n_header = iline + 1
                        break
                    iline += 1
                else:
                    raise ValueError(
                        "Could not find a sequence {n_cells} and '(' in file ({filename})"
                    )

        try:
            field = np.loadtxt(
                filename, dtype=tuple, skiprows=n_header, max_rows=n_cells
            )
            for i in range(n_cells):
                field[i, 0] = float(field[i, 0][1:])
                field[i, 1] = float(field[i, 1])
                field[i, 2] = float(field[i, 2][:-1])
            field = np.array(field).astype(float)

        except Exception as err:
            error_msg = f"Issue when reading {filename}"
            logger.error(error_msg)
            raise err(error_msg)

    return {
        "field": field,
        "name": meta_data["name"],
        "n_cells": n_cells,
        "n_header": n_header,
    }


def readOF(
    filename: str,
    n_cells: int | None = None,
    n_header: int | None = None,
    meta_data: dict | None = None,
) -> dict:
    """
    Read an OpenFOAM field outputted in ASCII format

    Parameters
    ----------
    filename: str
        Field filename
    n_cells : int | None
        Number of computational cells in the domain
    n_header : int | None
        Number of header lines
    meta_data : dict | None
        meta data dictionary
        If None, it is read from filename


    returns
    ----------
    data: dict
        Dictionary that contain info about the scalar field
            ======== =====================================================
            Key      Description (type)
            ======== =====================================================
            field    For nonuniform fields, array of size the number of cells (*np.ndarray*).
                     For uniform fields with a specified n_cells,
                          array of size the number of cells (*np.ndarray*).
                     For uniform fields, a scalar value (*float*)
            name     Name of the field (*str*)
            n_cells  Number of computational cells (*int*)
            n_header Number of header lines (*int*)
            ======== =====================================================
    """
    if meta_data is None:
        # Read meta data
        meta_data = read_meta_data(filename)
    if meta_data["type"] == "scalar":
        return readOFScal(filename=filename, meta_data=meta_data)
    if meta_data["type"] == "vector":
        return readOFVec(filename=filename, meta_data=meta_data)


def readSizeGroups(file):
    sizeGroup = {}
    f = open(file, "r")
    lines = f.readlines()
    f.close()
    begin = None
    for iline, line in enumerate(lines):
        if "sizeGroups" in line:
            begin = iline + 2
        if not begin is None and ");" in line:
            end = iline - 1
            break
    for line in lines[begin : end + 1]:
        tmp = line.split("{")
        name = tmp[0].strip()
        size = tmp[1].split(";")[0].split()[1]
        sizeGroup[name] = float(size)
    # Sort by size
    sizeGroup = dict(sorted(sizeGroup.items(), key=lambda item: item[1]))
    binGroup = {}
    groups = list(sizeGroup.keys())
    for igroup, group in enumerate(groups):
        if igroup == 0:
            bin_size = (
                sizeGroup[groups[igroup + 1]] - sizeGroup[groups[igroup]]
            )
        elif igroup == len(groups) - 1:
            bin_size = (
                sizeGroup[groups[igroup]] - sizeGroup[groups[igroup - 1]]
            )
        else:
            bin_size_p = (
                sizeGroup[groups[igroup + 1]] - sizeGroup[groups[igroup]]
            )
            bin_size_m = (
                sizeGroup[groups[igroup]] - sizeGroup[groups[igroup - 1]]
            )
            assert abs(bin_size_p - bin_size_m) < 1e-12
            bin_size = bin_size_m
        binGroup[group] = bin_size
    return sizeGroup, binGroup


def getCaseTimes(
    casePath: str, remove_zero: bool = False
) -> tuple[list[float], list[str]]:
    """
    Get list of all time folders from an OpenFOAM case

    Parameters
    ----------
    casePath: str
        Path to case folder
    remove_zero : bool
        Whether to remove zero from the time folder list

    returns
    ----------
    time_float_sorted: list[float]
        List of time folder values in ascending order
    time_str_sorted: list[str]
        List of time folder names in ascending order

    """
    # Read Time
    times_tmp = os.listdir(casePath)
    # remove non floats
    for i, entry in reversed(list(enumerate(times_tmp))):
        try:
            a = float(entry)
            if remove_zero:
                if abs(a) < 1e-12:
                    _ = times_tmp.pop(i)
        except ValueError:
            logger.debug(f"{entry} not a time folder, removing")
            a = times_tmp.pop(i)
            # print('removed ', a)
    time_float = [float(entry) for entry in times_tmp]
    time_str = [entry for entry in times_tmp]
    index_sort = np.argsort(time_float)
    time_float_sorted = [time_float[i] for i in list(index_sort)]
    time_str_sorted = [time_str[i] for i in list(index_sort)]

    return time_float_sorted, time_str_sorted


def getMeshTime(casePath: str) -> str:
    """
    Get the time at which the mesh was printed

    Parameters
    ----------
    casePath: str
        Path to case folder

    Returns
    ----------
    time_mesh: str
        The name of the time at which "meshFaceCentresXXX" was created
    """

    files_tmp = os.listdir(casePath)
    for entry in files_tmp:
        if "meshCellCentres" in entry:
            time_mesh = entry[16:-4]
            return time_mesh


def remove_comments(text: str) -> str:
    """
    Remove C++-style comments (// and /*) from the input and markers like #{ #}

    Parameters
    ----------
    text: str
        Raw input text containing comments

    Returns
    ----------
    text: str
        Text with all comments removed
    """

    # text = re.sub(
    #    r"/\*.*?\*/", "", text, flags=re.DOTALL
    # )  # Remove /* */ comments
    # text_unc = re.sub(r"//.*", "", text)  # Remove // comments

    text = re.sub(r"/\*.*?\*/", "", text, flags=re.DOTALL)
    text = re.sub(r"//.*", "", text)
    text = re.sub(r"#\{", "{", text)
    text = re.sub(r"#\};", "}", text)
    text = re.sub(r"#codeStream", "", text)
    return text


def tokenize(text: str) -> list[str]:
    """
    Add spaces around special characters (brace and semicolons) to make them separate tokens

    Parameters
    ----------
    text: str
        The cleaned (comment-free) OpenFOAM-style text.

    Returns:
    ----------
    token_list: list[str]
        List of tokens.
    """
    text = re.sub(r"([{}();])", r" \1 ", text)
    text = re.sub(r'"\s*\(\s*([^)]+?)\s*\)\s*"', r'"(\1)"', text)

    token_list = text.split()

    # print(token_list)
    # print(text)

    return token_list


def parse_tokens(tokens: list[str]) -> dict:
    """
    Parse OpenFOAM tokens into a nested Python dictionary.
    Special handling for `code { ... }` blocks to be stored as raw strings.

    Parameters
    ----------
    tokens: list[str]
        A list of tokens produced by `tokenize`.

    Returns
    ----------
    parsed: dict
        A nested dictionary that represents the OpenFOAM dictionary.
    """

    def parse_block(index: int) -> tuple:
        result = {}
        while index < len(tokens):
            token = tokens[index]
            if token == "}":
                return result, index + 1
            elif token == "{":
                raise SyntaxError("Unexpected '{'")

            key = token
            index += 1

            # key followed by dictionary
            if index < len(tokens) and tokens[index] == "{":
                index += 1
                if key == "code":
                    code_lines = []
                    while tokens[index] != "}":
                        code_lines.append(tokens[index])
                        index += 1
                    index += 1
                    if index < len(tokens) and tokens[index] == ";":
                        index += 1
                    result[key] = " ".join(code_lines).strip()
                else:
                    subdict, index = parse_block(index)
                    result[key] = subdict

            # key followed by list
            elif index < len(tokens) and tokens[index] == "(":
                index += 1

                # Peek to check if it's a dict-list (starts with '(' then '{')
                if tokens[index] == "(":
                    dictlist = {}
                    while tokens[index] != ")":
                        if tokens[index] != "(":
                            raise SyntaxError(
                                f"Expected '(' for label in dict-list, got {tokens[index]}"
                            )
                        # Read full label (e.g., "(gas and liquid)")
                        label_tokens = []
                        while tokens[index] != ")":
                            label_tokens.append(tokens[index])
                            index += 1
                        label_tokens.append(tokens[index])  # include ')'
                        index += 1
                        label = " ".join(label_tokens)

                        if tokens[index] != "{":
                            raise SyntaxError(
                                f"Expected '{{' after label {label}"
                            )
                        index += 1
                        subdict, index = parse_block(index)
                        dictlist[label] = subdict
                    index += 1  # skip final ')'
                    if index < len(tokens) and tokens[index] == ";":
                        index += 1
                    result[key] = dictlist
                else:
                    # Standard list
                    lst = []
                    while tokens[index] != ")":
                        lst.append(tokens[index])
                        index += 1
                    index += 1
                    if index < len(tokens) and tokens[index] == ";":
                        index += 1
                    result[key] = lst

            # key followed by scalar
            elif index < len(tokens):
                value = tokens[index]
                index += 1
                if index < len(tokens) and tokens[index] == ";":
                    index += 1
                result[key] = value

        return result, index

    parsed, _ = parse_block(0)
    return parsed


def read_openfoam_dict(filename: str) -> dict:
    """
    Parse OpenFOAM dictionary into a python dictionary

    Parameters
    ----------
    filename: str
        OpenFOAM dictionary filename

    Returns
    ----------
    dict_of: dict
        A Python dictionary representing the structure of the OpenFOAM dictionary.
    """
    with open(filename, "r+") as f:
        text = f.read()
    text = remove_comments(text)
    tokens = tokenize(text)
    foam_dict = parse_tokens(tokens)
    return foam_dict


def write_openfoam_dict(data: dict, filename: str, indent: int = 0) -> None:
    """
    Save a Python dictionary back to an OpenFOAM-style file.

    Parameters
    ----------
    d: dict
        Python dictionary to save
    filename: str
        The file that will contain the saved dictionary
    indent: int
        Number of indentation space
    """

    def write_block(f, key, value, indent=0):
        pad = " " * indent
        if isinstance(value, dict):
            f.write(f"{pad}{key}\n{pad}{{\n")
            for k, v in value.items():
                write_block(f, k, v, indent + 4)
            f.write(f"{pad}}}\n")
        elif isinstance(value, list):
            if all(isinstance(v, str) for v in value):
                f.write(f"{pad}{key}\n{pad}(\n")
                for v in value:
                    f.write(f"{pad}    {v}\n")
                f.write(f"{pad});\n")
            else:
                # assume list of numbers for OpenFOAM vectors
                joined = " ".join(value)
                f.write(f"{pad}{key}    ( {joined} );\n")
        else:
            f.write(f"{pad}{key}    {value};\n")

    with open(filename, "w") as f:
        # Write OpenFOAM header
        f.write(
            r"""/*--------------------------------*- C++ -*----------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     | Website:  https://openfoam.org
    \\  /    A nd           | Version:  9
     \\/     M anipulation  |
\*---------------------------------------------------------------------------*/
"""
        )
        # Write FoamFile block first
        foam_file = data.pop("FoamFile", None)
        if foam_file:
            write_block(f, "FoamFile", foam_file)
        f.write(
            "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n\n"
        )

        # Then write the rest of the blocks
        for key, value in data.items():
            write_block(f, key, value)
            f.write("\n")

        # Write OpenFOAM footer
        f.write(
            "// ************************************************************************* //\n"
        )



import logging

from prettyPlot.plotting import plt, pretty_labels, pretty_legend

logger = logging.getLogger(__name__)


def label_conv(input_string):
    if input_string.lower() == "width":
        return "width [mm]"
    elif input_string.lower() == "spacing":
        return "spacing [mm]"
    elif input_string.lower() == "height":
        return "height [mm]"
    elif (
        input_string.lower() == "co2_liq"
        or input_string.lower() == "co2.liquid"
    ):
        return r"$Y_{CO_2, liq}$"
    elif (
        input_string.lower() == "co_liq" or input_string.lower() == "co.liquid"
    ):
        return r"$Y_{CO, liq}$"
    elif (
        input_string.lower() == "h2_liq" or input_string.lower() == "h2.liquid"
    ):
        return r"$Y_{H_2, liq}$"
    elif (
        input_string.lower() == "co2_gas" or input_string.lower() == "co2.gas"
    ):
        return r"$Y_{CO_2, gas}$"
    elif input_string.lower() == "co_gas" or input_string.lower() == "co.gas":
        return r"$Y_{CO, gas}$"
    elif input_string.lower() == "h2_gas" or input_string.lower() == "h2.gas":
        return r"$Y_{H_2, gas}$"
    elif input_string.lower() == "kla_h2":
        return r"$KLA_{H_2}$"
    elif input_string.lower() == "kla_co":
        return r"$KLA_{CO}$"
    elif input_string.lower() == "kla_co2":
        return r"$KLA_{CO_2}$"
    elif input_string.lower() == "alpha.gas":
        return r"$\alpha_{gas}$"
    elif (
        input_string.lower() == "d.gas"
        or input_string.lower() == "d"
        or input_string.lower() == "bubblediam"
    ):
        return "Mean bubble diam [m]$"
    elif input_string.lower() == "y":
        return "y [m]"
    elif input_string.lower() == "t":
        return "t [s]"
    elif input_string.lower() == "gh":
        return "Gas holdup"
    elif input_string.lower() == "gh_height":
        return "Height-based gas holdup"
    else:
        logger.info(input_string)
        return input_string



import numpy as np
from prettyPlot.plotting import *

nsparg = [1, 2, 3, 4, 5, 6, 7, 8]
qoi_opt = [13.485, 14.276, 14.635, 15.692, 16.072, 17.657, 19.349, 20.712]

fig = plt.figure()
plt.plot(nsparg, qoi_opt, linewidth=3, color="k")
pretty_labels(
    r"N$_{\rm sparg}$", r"Optimal QOI [kg$^2$/kWh$^2$]", 20, fontname="Times"
)
plt.savefig("marginal_gain.png")
plt.savefig("marginal_gain.pdf")



import numpy as np
from prettyPlot.plotting import plt, pretty_labels, pretty_legend


def plotSTL(stl_file):
    from mpl_toolkits import mplot3d
    from stl import mesh

    # Create a new plot
    figure = plt.figure()
    axes = figure.add_subplot(projection="3d")

    # Load the STL files and add the vectors to the plot
    your_mesh = mesh.Mesh.from_file(stl_file)

    poly_collection = mplot3d.art3d.Poly3DCollection(your_mesh.vectors)
    poly_collection.set_color((0, 0, 0))
    axes.add_collection3d(poly_collection)

    min_x = np.amin(your_mesh.points[:, 0])
    max_x = np.amax(your_mesh.points[:, 0])
    min_y = np.amin(your_mesh.points[:, 1])
    max_y = np.amax(your_mesh.points[:, 1])
    min_z = np.amin(your_mesh.points[:, 2])
    max_z = np.amax(your_mesh.points[:, 2])

    amp = np.array([max_x - min_x, max_y - min_y, max_z - min_z])
    # 2D view
    if abs(amp[0]) < 1e-12:
        axes.view_init(0, 90)
    elif abs(amp[1]) < 1e-12:
        axes.view_init(0, 90)
        plt.yticks([])
    elif abs(amp[2]) < 1e-12:
        axes.view_init(0, 90)

    eps = np.amax(amp)
    axes.set_xlim3d(left=min_x - eps, right=max_x + eps)
    axes.set_ylim3d(min_y - eps, max_y + eps)
    axes.set_zlim3d(min_z - eps, max_z + eps)
    plt.locator_params(axis="x", nbins=4)
    plt.locator_params(axis="y", nbins=4)
    plt.locator_params(axis="z", nbins=4)

    return axes



import argparse


def initArgs():
    # CLI
    parser = argparse.ArgumentParser(
        description="Generate inhomogeneous boundary"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="plot on screen"
    )
    parser.add_argument(
        "-rc",
        "--r_const",
        type=float,
        metavar="",
        required=False,
        help="Constant radius value",
        default=0.1,
    )
    parser.add_argument(
        "-re",
        "--r_end",
        type=float,
        metavar="",
        required=False,
        help="End radius value",
        default=1,
    )
    parser.add_argument(
        "-pi",
        "--pore_in",
        type=float,
        metavar="",
        required=False,
        help="Pore diameter at center",
        default=3e-5,
    )
    parser.add_argument(
        "-po",
        "--pore_out",
        type=float,
        metavar="",
        required=False,
        help="Pore diameter at radius end",
        default=2e-5,
    )
    parser.add_argument(
        "-xc",
        "--xcent",
        type=float,
        metavar="",
        required=False,
        help="Column center x",
        default=0.0,
    )
    parser.add_argument(
        "-zc",
        "--zcent",
        type=float,
        metavar="",
        required=False,
        help="Column center z",
        default=0.0,
    )
    parser.add_argument(
        "-ds",
        "--diam_sparger",
        type=float,
        metavar="",
        required=False,
        help="Sparger diameter",
        default=0.15,
    )
    parser.add_argument(
        "-ugs",
        "--superf_vel",
        type=float,
        metavar="",
        required=False,
        help="Superficial velocity",
        default=0.01,
    )

    args = parser.parse_args()

    return args



import numpy as np

from bird.utilities.ofio import *

logger = logging.getLogger(__name__)


def read_field(
    case_folder: str,
    time_folder: str,
    field_name: str,
    n_cells: int | None = None,
    field_dict: dict = {},
) -> tuple[np.ndarray | float, dict]:
    """
    Read field at a given time and store it in dictionary for later reuse

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    field_name: str
        Name of the field file to read
    n_cells : int | None
        Number of cells in the domain
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    field : np.ndarray | float
        Field read
    field_dict : dict
        Dictionary of fields read
    """

    if not (field_name in field_dict) or field_dict[field_name] is None:
        # Read field if it had not been read before
        field_file = os.path.join(case_folder, time_folder, field_name)
        field = readOF(field_file, n_cells=n_cells)["field"]
        field_dict[field_name] = field
    else:
        # Get field from dict if it has been read before
        field = field_dict[field_name]

    return field, field_dict


def read_cell_volume(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    field_dict: dict = {},
) -> tuple[np.ndarray | float, dict]:
    """
    Read volume at a given time and store it in dictionary for later reuse

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    cell_volume : np.ndarray | float
        cell volume read from file
    field_dict : dict
        Dictionary of fields read
    """
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    try:
        cell_volume, field_dict = read_field(
            field_name="V", field_dict=field_dict, **kwargs_vol
        )
    except FileNotFoundError:
        error_msg = (
            f"Could not find {os.path.join(case_folder,volume_time,'V')}\n"
        )
        error_msg += "You can generate V with\n\t"
        error_msg += f"`postProcess -func writeCellVolumes -time {time_folder} -case {case_folder}`"
        logger.error(error_msg)
        raise FileNotFoundError(error_msg)

    return cell_volume, field_dict


def read_cell_centers(
    case_folder: str,
    cell_centers_file: str,
    field_dict: dict = {},
) -> tuple[np.ndarray, dict]:
    """
    Read volume at a given time and store it in dictionary for later reuse

    Parameters
    ----------
    case_folder: str
        Path to case folder
    cell_centers_file : str
        Filename of cell center data
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    cell_centers : np.ndarray
        cell centers read from file
    field_dict : dict
        Dictionary of fields read
    """

    if (
        not ("cell_centers" in field_dict)
        or field_dict["cell_centers"] is None
    ):
        try:
            cell_centers = readMesh(
                os.path.join(case_folder, cell_centers_file)
            )
            field_dict["cell_centers"] = cell_centers
        except FileNotFoundError:
            error_msg = f"Could not find {cell_centers_file}"
            error_msg += "You can generate it with\n\t"
            error_msg += f"`writeMeshObj -case {case_folder}`\n"
            time_float, time_str = getCaseTimes(case_folder)
            correct_ph = f"meshCellCentres_{time_str[0]}.obj"
            if not correct_path == cell_centers_file:
                error_msg += (
                    f"And adjust the cell center file path to {correct_path}"
                )
            logger.error(error_msg)
            raise FileNotFoundError(error_msg)
    else:
        cell_centers = field_dict["cell_centers"]

    return cell_centers, field_dict


def get_ind_liq(
    case_folder: str | None = None,
    time_folder: str | None = None,
    n_cells: int | None = None,
    field_dict: dict = {},
) -> tuple[np.ndarray | float, dict]:
    """
    Get indices of pure liquid cells (where alpha.liquid > 0.5)

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    field_name: str
        Name of the field file to read
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    ind_liq : np.ndarray | float
        indices of pure liquid cells
    field_dict : dict
        Dictionary of fields read
    """

    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }

    # Compute indices of pure liquid
    if not ("ind_liq" in field_dict) or field_dict["ind_liq"] is None:
        alpha_liq, field_dict = read_field(
            case_folder,
            time_folder,
            field_name="alpha.liquid",
            n_cells=n_cells,
            field_dict=field_dict,
        )
        ind_liq = np.argwhere(alpha_liq > 0.5)
        field_dict["ind_liq"] = ind_liq
    else:
        ind_liq = field_dict["ind_liq"]

    return ind_liq, field_dict


def get_ind_gas(
    case_folder: str | None = None,
    time_folder: str | None = None,
    n_cells: int | None = None,
    field_dict: dict = {},
) -> tuple[np.ndarray | float, dict]:
    """
    Get indices of pure gas cells (where alpha.liquid <= 0.5)

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    field_name: str
        Name of the field file to read
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    ind_gas : np.ndarray | float
        indices of pure gas cells
    field_dict : dict
        Dictionary of fields read
    """

    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }

    # Compute indices of pure liquid
    if not ("ind_gas" in field_dict) or field_dict["ind_gas"] is None:
        alpha_liq = read_field(
            case_folder,
            time_folder,
            field_name="alpha.liquid",
            n_cells=n_cells,
            field_dict=field_dict,
        )
        ind_gas = np.argwhere(alpha_liq <= 0.5)
        field_dict["ind_gas"] = ind_gas
    else:
        ind_gas = field_dict["ind_gas"]

    return ind_gas, field_dict


def compute_gas_holdup(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str = "0",
    field_dict: dict = {},
) -> tuple[float, dict]:
    """
    Calculate volume averaged gas hold up at a given time

    .. math::
       \frac{1}{V_{\rm tot}} \int_{V} (1-\alpha_{\rm liq}) dV

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    volume_time : str
        Time folder to read to get the cell volumes
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    gas_holdup: float
        Volume averaged gas holdup
    field_dict : dict
        Dictionary of fields read
    """

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    cell_volume, field_dict = read_cell_volume(
        field_dict=field_dict, **kwargs_vol
    )

    # Calculate
    gas_holdup = np.sum((1 - alpha_liq) * cell_volume) / np.sum(cell_volume)

    return gas_holdup, field_dict


def compute_superficial_velocity(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str = "0",
    direction: int = 1,
    cell_centers_file: str = "meshCellCentres_0.obj",
    field_dict: dict = {},
) -> tuple[float, dict]:
    """
    Calculate superficial velocity in a given direction at a given time

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    volume_time : str
        Time folder to read to get the cell volumes
    direction :  int
        Direction along which to calculate the superficial velocity
    cell_centers_file : str
        Filename of cell center data
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    sup_vel: float
        Superficial velocity
    field_dict : dict
        Dictionary of fields read
    """

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_gas, field_dict = read_field(
        field_name="alpha.gas", field_dict=field_dict, **kwargs
    )
    U_gas, field_dict = read_field(
        field_name="U.gas", field_dict=field_dict, **kwargs
    )
    U_gas = U_gas[:, direction]

    cell_volume, field_dict = read_cell_volume(
        field_dict=field_dict, **kwargs_vol
    )
    cell_centers, field_dict = read_cell_centers(
        case_folder=case_folder,
        cell_centers_file=cell_centers_file,
        field_dict=field_dict,
    )

    # Find all cells in the middle of the domain
    max_dir = np.amax(cell_centers[:, direction])
    min_dir = np.amin(cell_centers[:, direction])
    middle_dir = (max_dir + min_dir) / 2
    tol = (max_dir - min_dir) * 0.05
    ind_middle = np.argwhere(
        (cell_centers[:, direction] >= middle_dir - tol)
        & (cell_centers[:, direction] <= middle_dir + tol)
    )

    # Only compute in the middle
    if isinstance(alpha_gas, np.ndarray):
        alpha_gas = alpha_gas[ind_middle]
    if isinstance(cell_volume, np.ndarray):
        cell_volume = cell_volume[ind_middle]
    if isinstance(U_gas, np.ndarray):
        U_gas = U_gas[ind_middle]

    # Compute
    sup_vel = np.sum(U_gas * alpha_gas * cell_volume) / np.sum(cell_volume)

    return sup_vel, field_dict


def compute_ave_y_liq(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str = "0",
    spec_name: str = "CO2",
    field_dict={},
) -> tuple[float, dict]:
    """
    Calculate liquid volume averaged mass fraction of a species at a given time

    .. math::
       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} Y dV_{\rm liq}

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    volume_time : str
        Time folder to read to get the cell volumes
    spec_name : str
        Name of the species
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    liq_ave_y: float
        Liquid volume averaged mass fraction
    field_dict : dict
        Dictionary of fields read
    """

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    y_liq, field_dict = read_field(
        field_name=f"{spec_name}.liquid", field_dict=field_dict, **kwargs
    )
    ind_liq, field_dict = get_ind_liq(field_dict=field_dict, **kwargs)

    cell_volume, field_dict = read_cell_volume(
        field_dict=field_dict, **kwargs_vol
    )

    # Only compute over the liquid
    if isinstance(alpha_liq, np.ndarray):
        alpha_liq = alpha_liq[ind_liq]
    if isinstance(cell_volume, np.ndarray):
        cell_volume = cell_volume[ind_liq]
    if isinstance(y_liq, np.ndarray):
        y_liq = y_liq[ind_liq]

    # Calculate
    liq_ave_y = np.sum(alpha_liq * y_liq * cell_volume) / np.sum(
        alpha_liq * cell_volume
    )

    return liq_ave_y, field_dict


def compute_ave_conc_liq(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str = "0",
    spec_name: str = "CO2",
    mol_weight: float = 0.04401,
    rho_val: float | None = 1000,
    field_dict={},
) -> tuple[float, dict]:
    """
    Calculate liquid volume averaged concentration of a species at a given time

    .. math::
       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} \rho_{\rm liq} Y / W dV_{\rm liq}

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    volume_time : str
        Time folder to read to get the cell volumes
    spec_name : str
        Name of the species
    mol_weight : float
        Molecular weight of species (kg/mol)
    rho_val : float | None
        Constant density not available from time folder (kg/m3)
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    conc_ave: float
        Liquid volume averaged species concentration
    field_dict : dict
        Dictionary of fields read
    """

    logger.debug(
        f"Computing concentration for {spec_name} with molecular weight {mol_weight:.4g} kg/mol"
    )
    if rho_val is not None:
        logger.debug(f"Assuming liquid density {rho_val} kg/m3")

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    y_liq, field_dict = read_field(
        field_name=f"{spec_name}.liquid", field_dict=field_dict, **kwargs
    )
    ind_liq, field_dict = get_ind_liq(field_dict=field_dict, **kwargs)

    cell_volume, field_dict = read_cell_volume(
        field_dict=field_dict, **kwargs_vol
    )

    # Density of liquid is not always printed (special case)
    if not ("rho_liq" in field_dict) or field_dict["rho_liq"] is None:
        if rho_val is not None:
            rho_liq = rho_val
            field_dict["rho_liq"] = rho_val
        else:
            rho_liq_file = os.path.join(case_folder, time_folder, "rhom")
            rho_liq = readOFScal(rho_liq_file, n_cells)["field"]
            field_dict["rho_liq"] = rho_liq
    else:
        rho_liq = field_dict["rho_liq"]

    # Only compute over the liquid
    if isinstance(y_liq, np.ndarray):
        y_liq = y_liq[ind_liq]
    if isinstance(alpha_liq, np.ndarray):
        alpha_liq = alpha_liq[ind_liq]
    if isinstance(alpha_liq, np.ndarray):
        cell_volume = cell_volume[ind_liq]
    if isinstance(rho_liq, np.ndarray):
        rho_liq = rho_liq[ind_liq]

    conc_loc = rho_liq * y_liq / mol_weight

    conc_ave = np.sum(conc_loc * alpha_liq * cell_volume) / np.sum(
        alpha_liq * cell_volume
    )

    return conc_ave, field_dict


def compute_ave_bubble_diam(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str = "0",
    field_dict={},
) -> tuple[float, dict]:
    """
    Calculate averaged bubble diameter over the liquid volume

    .. math::
       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} D dV

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain
    volume_time : str
        Time folder to read to get the cell volumes
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    diam: float
        Volume averaged gas holdup
    field_dict : dict
        Dictionary of fields read
    """

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    d_gas, field_dict = read_field(
        field_name="d.gas", field_dict=field_dict, **kwargs
    )
    ind_liq, field_dict = get_ind_liq(field_dict=field_dict, **kwargs)

    cell_volume, field_dict = read_cell_volume(
        field_dict=field_dict, **kwargs_vol
    )

    # Only compute over the liquid
    if isinstance(d_gas, np.ndarray):
        d_gas = d_gas[ind_liq]
    if isinstance(alpha_liq, np.ndarray):
        alpha_liq = alpha_liq[ind_liq]
    if isinstance(alpha_liq, np.ndarray):
        cell_volume = cell_volume[ind_liq]

    # Calculate
    diam = np.sum(d_gas * alpha_liq * cell_volume) / np.sum(
        alpha_liq * cell_volume
    )

    return diam, field_dict



import logging
import os

import numpy as np

from bird.meshing._mesh_tools import *
from bird.utilities.parser import parse_json, parse_yaml

logger = logging.getLogger(__name__)


def assemble_geom(input_file, topo_file):
    if input_file.endswith(".yaml"):
        inpt_dict = parse_yaml(input_file)
    elif input_file.endswith(".json"):
        inpt_dict = parse_json(input_file)
    else:
        raise ValueError(f"unknown input file ({input_file}) extension")
    if topo_file.endswith(".yaml"):
        topo_dict = parse_yaml(topo_file)
    elif topo_file.endswith(".json"):
        topo_dict = parse_json(topo_file)
    else:
        raise ValueError(f"unknown topo file ({topo_file}) extension")

    # ~~~~ Define dimensions based on input
    r_dimensions_name = list(inpt_dict["Geometry"]["Radial"].keys())
    r_dimensions = [
        float(inpt_dict["Geometry"]["Radial"][dim])
        for dim in r_dimensions_name
    ]

    l_dimensions_name = list(inpt_dict["Geometry"]["Longitudinal"].keys())
    l_dimensions = [
        float(inpt_dict["Geometry"]["Longitudinal"][dim])
        for dim in l_dimensions_name
    ]

    # Order the dimensions of R and L
    r_coord = merge_and_sort(coord_list=r_dimensions, reverse_coord=False)
    l_coord = merge_and_sort(coord_list=l_dimensions, reverse_coord=True)
    dimension_dict = {"R": r_coord, "L": l_coord}

    # Define blocks that will be walls
    wall_dict = make_walls_from_topo(topo_dict)
    bound_dict = make_bound_from_topo(topo_dict)

    return {**wall_dict, **bound_dict, **dimension_dict}


def assemble_mesh(input_file, geom_dict):
    if input_file.endswith(".yaml"):
        inpt_dict = parse_yaml(input_file)
    elif input_file.endswith(".json"):
        inpt_dict = parse_json(input_file)
    else:
        raise ValueError(f"unknown input file ({input_file}) extension")
    R = geom_dict["R"]
    L = geom_dict["L"]
    N1 = len(R)
    N2 = len(L) - 1
    CW = []
    mCW = []
    C1 = []
    mC1 = []
    C2 = []
    mC2 = []
    for rval in R:
        CW.append(rval * 0.5)
        mCW.append(-rval * 0.5)
        C1.append(rval * np.cos(np.pi / 4))
        mC1.append(-rval * np.cos(np.pi / 4))
        C2.append(rval * np.sin(np.pi / 4))
        mC2.append(-rval * np.sin(np.pi / 4))

    NRSmallest = int(inpt_dict["Meshing"]["NRSmallest"])
    NVertSmallest = int(inpt_dict["Meshing"]["NVertSmallest"])
    NS = []
    NR = []
    NVert = []
    gradR_l = []
    gradR_r = []
    gradR = []
    gradVert = []

    # Radial meshing
    rad_len_block = np.zeros(len(R))
    rad_len_block[0] = R[0] / 2
    for i in range(len(R) - 1):
        rad_len_block[i + 1] = abs(R[i + 1] - R[i])

    try:
        iSmallest = int(inpt_dict["Meshing"]["iRSmallest"])
    except KeyError:
        iSmallest = np.argmin(rad_len_block)

    i_smallest_rad = iSmallest
    smallestRBlockSize = rad_len_block[i_smallest_rad]

    for irad_block in range(len(rad_len_block)):
        logger.debug(
            f"\tRadial block {irad_block+1} has size {rad_len_block[irad_block]}mm"
        )
    logger.debug(
        f"Smallest radial block {i_smallest_rad+1} with size {smallestRBlockSize}mm"
    )
    logger.debug(
        f"Radial mesh size in the smallest block will be {smallestRBlockSize}/{NRSmallest} = {smallestRBlockSize/NRSmallest:.2f}mm"
    )

    NR = [0 for i in range(len(R))]
    NR[iSmallest] = NRSmallest
    for i in range(len(R)):
        if not i == iSmallest:
            NR[i] = max(
                int(
                    round(
                        NRSmallest * abs(rad_len_block[i]) / smallestRBlockSize
                    )
                ),
                1,
            )
    NS = [NR[0] * 2]
    # Now figure out grading of each block
    for ir in range(len(R)):
        gradR_l.append(1.0)
        gradR_r.append(1.0)
        gradR.append(1.0)

    # Vertical meshing
    vert_len_block = np.array(
        [abs(L[i + 1] - L[i]) for i in range(len(L) - 1)]
    )

    try:
        iSmallest = int(inpt_dict["Meshing"]["iVertSmallest"])
    except KeyError:
        iSmallest = np.argmin(vert_len_block)

    i_smallest_vert = iSmallest
    smallestVertBlockSize = vert_len_block[i_smallest_vert]

    for ivert_block in range(len(vert_len_block)):
        logger.debug(
            f"\tVertical block {ivert_block+1} has size {vert_len_block[ivert_block]}mm"
        )
    logger.debug(
        f"Smallest vertical block {i_smallest_vert+1} with size {smallestVertBlockSize}mm"
    )
    logger.debug(
        f"Vertical mesh size in the smallest block will be {smallestVertBlockSize}/{NVertSmallest} = {smallestVertBlockSize/NVertSmallest:.2f}mm"
    )

    NVert = [0] * (len(L) - 1)
    NVert[iSmallest] = NVertSmallest
    for i in range(len(L) - 1):
        if not i == iSmallest:
            NVert[i] = max(
                int(
                    round(
                        NVert[iSmallest]
                        * abs(L[i + 1] - L[i])
                        / smallestVertBlockSize
                    )
                ),
                1,
            )
    for il in range(len(L) - 1):
        gradVert.append(1.0)

    # Mesh stretching
    try:
        verticalCoarseningProperties = inpt_dict["Meshing"][
            "verticalCoarsening"
        ]
        do_verticalCoarsening = True
    except KeyError:
        do_verticalCoarsening = False
    try:
        radialCoarseningProperties = inpt_dict["Meshing"]["radialCoarsening"]
        do_radialCoarsening = True
    except KeyError:
        do_radialCoarsening = False

    if do_verticalCoarsening:
        NVert, gradVert, minCellVert, maxCellVert = verticalCoarsening(
            ratio_properties=verticalCoarseningProperties,
            ref_block=i_smallest_vert,
            NVert=NVert,
            L=L,
            gradVert=gradVert,
            smooth=True,
        )
    else:
        block_length = [abs(L[i] - L[i + 1]) for i in range(len(NVert))]
        block_cell_length = [
            block_length[i] / NVert[i] for i in range(len(NVert))
        ]
        minCellVert = np.amin(block_cell_length)
        maxCellVert = np.amax(block_cell_length)
    if do_radialCoarsening:
        NR, gradR, minCellR, maxCellR = radialCoarsening(
            ratio_properties=radialCoarseningProperties,
            ref_block=i_smallest_rad,
            NR=NR,
            R=R,
            gradR=gradR,
            smooth=True,
        )
    else:
        block_length = [R[0] / 2] + [
            abs(R[i] - R[i + 1]) for i in range(len(R) - 1)
        ]
        block_cell_length = [block_length[i] / NR[i] for i in range(len(NR))]
        minCellR = np.amin(block_cell_length)
        maxCellR = np.amax(block_cell_length)

    logger.info("Vertical mesh:")
    logger.info(f"\tTotal NVert {sum(NVert)}")
    logger.info(f"\tNVert {NVert}")
    logger.info(f"\tsize min {minCellVert:.2f}mm max {maxCellVert:.2f}mm")
    logger.info("Radial mesh:")
    logger.info(f"\tTotal NR {sum(NR)}")
    logger.info(f"\tNR {NR}")
    logger.info(f"\tsize min {minCellR:.2f}mm max {maxCellR:.2f}mm")

    return {
        "NR": NR,
        "NS": NS,
        "NVert": NVert,
        "gradR": gradR,
        "gradVert": gradVert,
        "gradR_l": gradR_l,
        "gradR_r": gradR_r,
        "CW": CW,
        "mCW": mCW,
        "C1": C1,
        "mC1": mC1,
        "C2": C2,
        "mC2": mC2,
    }


def writeBlockMeshDict(out_folder, geom_dict, mesh_dict):
    outfile = os.path.join(out_folder, "blockMeshDict")

    R = geom_dict["R"]
    L = geom_dict["L"]
    r_wall = geom_dict["r_wall"]
    l_wall = geom_dict["l_wall"]
    BoundaryNames = geom_dict["names"]
    BoundaryType = geom_dict["types"]
    BoundaryRmin = geom_dict["rmin"]
    BoundaryRmax = geom_dict["rmax"]
    BoundaryLmin = geom_dict["lmin"]
    BoundaryLmax = geom_dict["lmax"]

    NR = mesh_dict["NR"]
    NS = mesh_dict["NS"]
    NVert = mesh_dict["NVert"]
    gradR = mesh_dict["gradR"]
    gradR_l = mesh_dict["gradR_l"]
    gradR_r = mesh_dict["gradR_r"]
    gradVert = mesh_dict["gradVert"]
    CW = mesh_dict["CW"]
    mCW = mesh_dict["mCW"]
    C1 = mesh_dict["C1"]
    mC1 = mesh_dict["mC1"]
    C2 = mesh_dict["C2"]
    mC2 = mesh_dict["mC2"]

    N1 = len(R)
    N2 = len(L) - 1

    # ~~~~ Write species Dict
    fw = open(outfile, "w+")
    # Write Header
    fw.write("FoamFile\n")
    fw.write("{\n")
    fw.write("    version     2.0;\n")
    fw.write("    format      ascii;\n")
    fw.write("    class       dictionary;\n")
    fw.write("    object      blockMeshDict;\n")
    fw.write("}\n")
    fw.write("\n")
    # fw.write('convertToMeters 0.001;\n')
    # Write all radii
    counter = 1
    for rval in R:
        fw.write("R" + str(counter) + " " + str(rval) + ";\n")
        counter = counter + 1
    fw.write("\n")
    # Write all minus radii
    counter = 1
    for rval in R:
        fw.write("mR" + str(counter) + " " + str(-rval) + ";\n")
        counter = counter + 1
    fw.write("\n")
    # Write all Length
    counter = 1
    for lval in L:
        fw.write("L" + str(counter) + " " + str(lval) + ";\n")
        counter = counter + 1
    fw.write("\n")
    # Write all C
    counter = 1
    for rval in R:
        fw.write("CW" + str(counter) + " " + str(CW[counter - 1]) + ";\n")
        fw.write("mCW" + str(counter) + " " + str(mCW[counter - 1]) + ";\n")
        fw.write("C1" + str(counter) + " " + str(C1[counter - 1]) + ";\n")
        fw.write("mC1" + str(counter) + " " + str(mC1[counter - 1]) + ";\n")
        fw.write("C2" + str(counter) + " " + str(C2[counter - 1]) + ";\n")
        fw.write("mC2" + str(counter) + " " + str(mC2[counter - 1]) + ";\n")
        fw.write("\n")
        counter = counter + 1

    # Write all Ngrid
    counter = 1
    for nR in NR:
        fw.write("NR" + str(counter) + " " + str(NR[counter - 1]) + ";\n")
        counter = counter + 1
    fw.write("\n")
    counter = 1
    for nS in NS:
        fw.write("NS" + str(counter) + " " + str(NS[counter - 1]) + ";\n")
        counter = counter + 1
    fw.write("\n")
    counter = 1
    for nVert in NVert:
        fw.write(
            "NVert" + str(counter) + " " + str(NVert[counter - 1]) + ";\n"
        )
        counter = counter + 1
    fw.write("\n")

    # ~~~~ Write vertices
    fw.write("vertices\n")
    fw.write("(\n")
    # Write the squares first
    counter = 0
    for i in range(len(L)):
        fw.write(f"     ($mCW1  $mCW1  $L{i+1})// {counter}\n")
        fw.write(f"     ( $CW1  $mCW1  $L{i+1})\n")
        fw.write(f"     ( $CW1   $CW1  $L{i+1})\n")
        fw.write(f"     ($mCW1   $CW1  $L{i+1})\n")
        fw.write("\n")
        counter = counter + 4
    # Write the circles then
    for ir in range(len(R)):
        for il in range(len(L)):
            fw.write(f"    ($mC1{ir+1}   $mC2{ir+1}   $L{il+1})// {counter}\n")
            fw.write(f"    ( $C1{ir+1}   $mC2{ir+1}   $L{il+1})\n")
            fw.write(f"    ( $C1{ir+1}   $C2{ir+1}    $L{il+1})\n")
            fw.write(f"    ($mC1{ir+1}   $C2{ir+1}    $L{il+1})\n")
            fw.write("\n")
            counter = counter + 4
    fw.write(");\n")
    fw.write("\n")

    # ~~~~ Write blocks
    fw.write("blocks\n")
    fw.write("(\n")
    # Write the squares first
    for i in range(N2):
        gradingVert = gradVert[i]
        # Am I a wall
        iwall = is_wall(l_wall, r_wall, -1, i)
        if iwall == 1:
            fw.write("//")

        i1 = int((i + 1) * 4)
        i2 = int(i * 4)
        fw.write(
            f"     hex ({i1} {i1+1} {i1+2} {i1+3} {i2} {i2+1} {i2+2} {i2+3})"
        )
        fw.write(
            f" ($NS1 $NS1 $NVert{i+1}) simpleGrading (1 1 {gradingVert})\n"
        )
    fw.write("\n")
    # Write the squares then
    for ir in range(N1):
        for il in range(N2):
            # gradingVert = 1
            # gradingR = 1
            # if il==N2-1:
            #    gradingVert = outletGrading
            gradingR_l = gradR_l[ir]
            gradingR_r = gradR_r[ir]
            gradingVert = gradVert[il]
            gradingR = gradR[ir]
            # Am I a wall
            iwall = is_wall(l_wall, r_wall, ir, il)
            # bottom right corner
            i1 = int(4 * (N2 + 1) * ir + 4 * (il + 1))
            # bottom left corner
            i2 = int(4 * (N2 + 1) * (ir + 1) + 4 * (il + 1))
            # top right corner
            i3 = int(4 * (N2 + 1) * ir + 4 * (il))
            # top left corner
            i4 = int(4 * (N2 + 1) * (ir + 1) + 4 * (il))
            # outlet
            if iwall == 1:
                fw.write("//")
            fw.write(
                f"     hex ({i1} {i2} {i2+1} {i1+1} {i3} {i4} {i4+1} {i3+1})"
            )
            fw.write(
                f" ($NR{ir+1} $NS1  $NVert{il+1}) simpleGrading ({gradingR} 1 {gradingVert})\n"
            )
            if iwall == 1:
                fw.write("//")
            fw.write(
                f"     hex ({i1+1} {i2+1} {i2+2} {i1+2} {i3+1} {i4+1} {i4+2} {i3+2})"
            )
            fw.write(
                f" ($NR{ir+1} $NS1  $NVert{il+1}) simpleGrading ({gradingR} 1 {gradingVert})\n"
            )
            if iwall == 1:
                fw.write("//")
            fw.write(
                f"     hex ({i1+2} {i2+2} {i2+3} {i1+3} {i3+2} {i4+2} {i4+3} {i3+3})"
            )
            fw.write(
                f" ($NR{ir+1} $NS1  $NVert{il+1}) simpleGrading ({gradingR} 1 {gradingVert})\n"
            )
            if iwall == 1:
                fw.write("//")
            fw.write(
                f"     hex ({i1+3} {i2+3} {i2} {i1} {i3+3} {i4+3} {i4} {i3})"
            )
            fw.write(
                f" ($NR{ir+1} $NS1  $NVert{il+1}) simpleGrading ({gradingR} 1 {gradingVert})\n"
            )
            fw.write("\n")
    fw.write(");\n")
    fw.write("\n")

    # ~~~~ Write edges
    fw.write("edges\n")
    fw.write("(\n")
    ind = len(L) * 4
    for ir in range(len(R)):
        for il in range(len(L)):
            # Edges should be removed if they are surrounded by walls
            iwall1 = is_wall(l_wall, r_wall, ir, il)
            iwall2 = is_wall(l_wall, r_wall, ir + 1, il)
            iwall3 = is_wall(l_wall, r_wall, ir, il - 1)
            iwall4 = is_wall(l_wall, r_wall, ir + 1, il - 1)
            sumwall = iwall1 + iwall2 + iwall3 + iwall4
            comment = 0
            if (
                sumwall == 4
                or (il == 0 and sumwall == 2)
                or (il == len(L) - 1 and sumwall == 2)
                or (ir == len(R) - 1 and sumwall == 2)
                or (ir == len(R) - 1 and il == 0 and sumwall == 1)
            ):
                comment = 1

            if comment == 1:
                fw.write("//")
            fw.write(f"    arc {ind} {ind+1} (0    $mR{ir+1} $L{il+1})\n")
            if comment == 1:
                fw.write("//")
            fw.write(f"    arc {ind+1} {ind+2} ($R{ir+1} 0 $L{il+1})\n")
            if comment == 1:
                fw.write("//")
            fw.write(f"    arc {ind+2} {ind+3} (0 $R{ir+1} $L{il+1})\n")
            if comment == 1:
                fw.write("//")
            fw.write(f"    arc {ind+3} {ind} ($mR{ir+1} 0 $L{il+1})\n")
            ind = ind + 4
            fw.write("\n")
    fw.write(");\n")
    fw.write("\n")

    # ~~~~ Write boundary
    fw.write("boundary\n")
    fw.write("(\n")

    for i, name in enumerate(BoundaryNames):
        fw.write(f"    {name}\n")
        fw.write("    {\n")
        if name.startswith("wall"):
            fw.write("        " + "type wall;\n")
        else:
            fw.write("        " + "type patch;\n")
        fw.write("        faces\n")
        fw.write("        (\n")

        for ibound in range(len(BoundaryType[i])):
            boundType = BoundaryType[i][ibound]
            if boundType == "lateral":
                rminInd = BoundaryRmin[i][ibound]
                rmaxInd = BoundaryRmax[i][ibound]
                lInd = BoundaryLmin[i][ibound]
                i1 = rminInd * (4 * (N2 + 1)) + 4 * lInd  # bottom
                i2 = i1 - 4  # top
                fw.write(f"            ( {i1} {i1+1} {i2+1} {i2})\n")
                fw.write(f"            ( {i1+1} {i1+2} {i2+2} {i2+1})\n")
                fw.write(f"            ( {i1+2} {i1+3} {i2+3} {i2+2})\n")
                fw.write(f"            ( {i1+3} {i1} {i2} {i2+3})\n")
                fw.write("\n")

            elif boundType == "top":
                lminInd = BoundaryLmin[i][ibound]
                lmaxInd = BoundaryLmax[i][ibound]
                rInd = BoundaryRmin[i][ibound]
                if rInd > 0:
                    i1 = 4 * (N2 + 1) * (rInd - 1) + 4 * lminInd  # right
                    i2 = i1 + 4 * (N2 + 1)  # left
                    fw.write(f"            ( {i1} {i2} {i2+1} {i1+1})\n")
                    fw.write(f"            ( {i1+1} {i2+1} {i1+2} {i2+2})\n")
                    fw.write(f"            ( {i1+2} {i2+2} {i2+3} {i1+3})\n")
                    fw.write(f"            ( {i1+3} {i2+3} {i2} {i1})\n")
                else:
                    i1 = lminInd * 4
                    fw.write(f"            ( {i1} {i1+1} {i1+2} {i1+3})\n")
                fw.write("\n")

            elif boundType == "bottom":
                lminInd = BoundaryLmin[i][ibound]
                lmaxInd = BoundaryLmax[i][ibound]
                rInd = BoundaryRmin[i][ibound]
                i1 = 4 * (N2 + 1) * (rInd - 1) + 4 * lminInd  # right
                i2 = i1 + 4 * (N2 + 1)  # left
                if rInd > 0:
                    fw.write(f"            ( {i1} {i2} {i2+1} {i1+1})\n")
                    fw.write(f"            ( {i1+1} {i2+1} {i1+2} {i2+2})\n")
                    fw.write(f"            ( {i1+2} {i2+2} {i2+3} {i1+3})\n")
                    fw.write(f"            ( {i1+3} {i2+3} {i2} {i1})\n")
                else:
                    i1 = lminInd * 4
                    fw.write(f"            ( {i1} {i1+1} {i1+2} {i1+3})\n")
                fw.write("\n")

        fw.write("        " + ");\n")
        fw.write("    " + "}\n")

    fw.write(");\n")
    fw.write("\n")

    # ~~~~ Write mergePatchPairs
    fw.write("mergePatchPairs\n")
    fw.write("(\n")
    fw.write(");\n")

    fw.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    fw.write("defaultPatch\n{type wall;}\n\n")

    fw.close()


def main(input_file, topo_file, output_folder):
    geom_dict = assemble_geom(input_file, topo_file)
    mesh_dict = assemble_mesh(input_file, geom_dict)
    writeBlockMeshDict(output_folder, geom_dict, mesh_dict)


if __name__ == "__main__":
    main()



import json
import logging

import numpy as np

from bird.meshing.block_rect_mesh import from_block_rect_to_seg
from bird.preprocess.stl_patch.stl_mesh import STLMesh

logger = logging.getLogger(__name__)


def make_polygon(rad, nvert, center, normal_dir):
    logger.debug(
        f"\tMaking polygon at ({center[0]:.4g}, {center[1]:.4g}, {center[2]:.4g})"
    )
    theta = 2 * np.pi / nvert
    vertices = np.zeros((nvert, 3))
    t1dir = (normal_dir + 1) % 3
    t2dir = (normal_dir + 2) % 3
    for i in range(nvert):
        vertices[i, t2dir] = rad * np.cos(theta * i + (np.pi / 2 - theta / 2))
        vertices[i, t1dir] = rad * np.sin(theta * i + (np.pi / 2 - theta / 2))

    stl_mesh = STLMesh(vertices=vertices, planar=True, normal_dir=normal_dir)
    stl_mesh.translate(center)

    return stl_mesh


def make_rectangle(w, h, center, normal_dir):
    logger.debug(
        f"\tMaking rectangle at ({center[0]:.4g}, {center[1]:.4g}, {center[2]:.4g})"
    )
    # Define vertices
    t1dir = (normal_dir + 1) % 3
    t2dir = (normal_dir + 2) % 3
    vertices = np.zeros((4, 3))
    vertices[0, t2dir] = -w / 2
    vertices[0, t1dir] = h / 2
    vertices[1, t2dir] = w / 2
    vertices[1, t1dir] = h / 2
    vertices[2, t2dir] = w / 2
    vertices[2, t1dir] = -h / 2
    vertices[3, t2dir] = -w / 2
    vertices[3, t1dir] = -h / 2

    stl_mesh = STLMesh(vertices=vertices, planar=True, normal_dir=normal_dir)
    stl_mesh.translate(center)

    return stl_mesh


def make_circle(radius, center, normal_dir, npts=3):
    logger.debug(
        f"\tMaking circle at ({center[0]:.4g}, {center[1]:.4g}, {center[2]:.4g})"
    )
    vertices = np.zeros((npts + 1, 3))
    t1dir = (normal_dir + 1) % 3
    t2dir = (normal_dir + 2) % 3
    for i in range(npts):
        ang = i * 2.0 * np.pi / npts
        vertices[i + 1, t2dir] = radius * np.cos(ang)
        vertices[i + 1, t1dir] = radius * np.sin(ang)

    stl_mesh = STLMesh(vertices=vertices, planar=True, normal_dir=normal_dir)
    stl_mesh.translate(center)

    return stl_mesh


def make_spider(centerRad, nArms, widthArms, lengthArms, center, normal_dir):
    globalArea = 0
    if nArms < 2:
        error_msg = f"nArms ({nArms}) must be >= 2"
        logger.error(error_msg)
        raise ValueError(error_msg)
    if nArms == 2:
        nVertPol = 4
    if nArms > 2:
        nVertPol = nArms
    center_mesh = make_polygon(
        rad=centerRad, nvert=nVertPol, center=(0, 0, 0), normal_dir=normal_dir
    )
    vertices = center_mesh.vertices
    maxWidth = np.linalg.norm((vertices[1, :] - vertices[0, :]))
    if widthArms > maxWidth:
        error_msg = "arm width will make arms overlap"
        error_msg += "\nEither increase center radius or reduce arm width"
        logger.error(error_msg)
        raise ValueError(error_msg)

    arm_mesh_list = []
    for i in range(nArms):
        if nArms > 2:
            if i < nArms - 1:
                indp = i + 1
                indm = i
            else:
                indp = 0
                indm = i
        if nArms == 2:
            if i == 0:
                indp = 1
                indm = 0
            else:
                indp = 3
                indm = 2
        arm_mesh = make_rectangle(
            w=widthArms, h=lengthArms, center=(0, 0, 0), normal_dir=normal_dir
        )
        side = vertices[indp, :] - vertices[indm, :]
        if normal_dir == 1:
            angle = np.arccos(np.dot(side, [1, 0, 0]) / np.linalg.norm(side))
            if side[2] <= 0:
                angle *= -1
        else:
            raise NotImplementedError
        arm_mesh.rotate(angle, normal_dir=normal_dir)
        midSide = (vertices[indp, :] + vertices[indm, :]) / 2
        trans = midSide * (1 + lengthArms / (2 * np.linalg.norm(midSide)))
        arm_mesh.translate(trans)
        arm_mesh_list.append(arm_mesh)

    spider_mesh = STLMesh()
    spider_mesh.from_mesh_list([center_mesh] + arm_mesh_list)
    spider_mesh.translate(center)
    spider_mesh.planar = True
    spider_mesh.normal_dir = normal_dir

    return spider_mesh


def get_cent_from_branch(patch, input_dict):
    geom_dict = from_block_rect_to_seg(input_dict["Geometry"], rescale=False)
    segment = geom_dict["segments"][patch["branch_id"]]
    cent = segment["start"] + patch["frac_space"] * segment["conn"]
    ndir = patch["normal_dir"]
    if patch["block_pos"].lower() == "bottom":
        cent[ndir] -= geom_dict["blocksize"][ndir] / 2
    elif patch["block_pos"].lower() == "top":
        cent[ndir] += geom_dict["blocksize"][ndir] / 2
    return cent


def create_boundary_patch_list(input_dict, boundary_name):
    patch_mesh_list = []
    for patch in input_dict[boundary_name]:
        if "centx" in patch and "centy" in patch and "centz" in patch:
            cent = [patch["centx"], patch["centy"], patch["centz"]]
        elif "branch_id" in patch:
            cent = get_cent_from_branch(patch, input_dict)
        else:
            raise NotImplementedError
        if patch["type"].lower() == "circle":
            patch_mesh = make_circle(
                radius=patch["radius"],
                center=(cent[0], cent[1], cent[2]),
                normal_dir=patch["normal_dir"],
                npts=patch["nelements"],
            )
        elif patch["type"].lower() == "spider":
            patch_mesh = make_spider(
                centerRad=patch["centerRad"],
                nArms=patch["nArms"],
                widthArms=patch["widthArms"],
                lengthArms=patch["lengthArms"],
                center=(cent[0], cent[1], cent[2]),
                normal_dir=patch["normal_dir"],
            )
        elif patch["type"].lower() == "polygon":
            patch_mesh = make_polygon(
                rad=patch["radius"],
                nvert=patch["npts"],
                center=(cent[0], cent[1], cent[2]),
                normal_dir=patch["normal_dir"],
            )
        elif patch["type"].lower() == "rectangle":
            patch_mesh = make_rectangle(
                w=patch["width"],
                h=patch["height"],
                center=(cent[0], cent[1], cent[2]),
                normal_dir=patch["normal_dir"],
            )
        else:
            raise NotImplementedError
        patch_mesh_list.append(patch_mesh)

    return patch_mesh_list



import csv
import os
import pickle as pkl

import numpy as np


def get_config_result(study_fold: str = ".") -> None:
    """
    Read the configs.pkl and results.pkl files from a study
    Saves the configuration in Xdata_{study_fold}.csv file
    Save the qoi and qoi_error in ydata_{study_fold}.csv file

    Parameters
    ----------
    study_fold : str
        Folder that contains the study results

    Returns
    ----------
    None

    """
    # Read results
    with open(os.path.join(study_fold, "results.pkl"), "rb") as f:
        results = pkl.load(f)
    with open(os.path.join(study_fold, "configs.pkl"), "rb") as f:
        configs = pkl.load(f)

    Xdata = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], np.int64)
    count = 0

    # Save data into CSV files
    xfname = os.path.join(study_fold, f"Xdata.csv")
    yfname = os.path.join(study_fold, f"ydata.csv")
    with open(xfname, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        for sims in results:
            b0 = configs[sims][0]
            b1 = configs[sims][1]
            b2 = configs[sims][2]
            raw_data = np.concatenate((b0, b1, b2), axis=None)
            writer.writerow(raw_data)

    with open(yfname, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        for sims in results:
            q0 = results[sims]["qoi"]
            q1 = results[sims]["qoi_err"]
            y_data = np.concatenate((q0, q1), axis=None)
            writer.writerow(y_data)


if __name__ == "__main__":
    studies = {
        "study_scaleup_0_4vvm_3000W": r"608$m^3$ 0.4vvm 3000W",
        "study_scaleup_0_1vvm_6000W": r"608$m^3$ 0.1vvm 6000W",
        "study_0_4vvm_1W": r"0.00361$m^3$ 0.4vvm 1W",
    }
    for study in studies:
        get_config_result(study)



"""Bio reactor design"""

import os

from bird.version import __version__

BIRD_DIR = os.path.dirname(os.path.realpath(__file__))
BIRD_CASE_DIR = os.path.join(BIRD_DIR, "../tutorial_cases")
BIRD_CONST_DIR = os.path.join(BIRD_DIR, "constants")
BIRD_MESH_DIR = os.path.join(BIRD_DIR, "meshing")
BIRD_POST_DIR = os.path.join(BIRD_DIR, "postprocess")
BIRD_PRE_DIR = os.path.join(BIRD_DIR, "preprocess")
BIRD_BLOCK_CYL_MESH_TEMP_DIR = os.path.join(
    BIRD_MESH_DIR, "block_cyl_mesh_templates"
)
BIRD_BLOCK_CYL_CASE_TEMP_DIR = os.path.join(
    BIRD_MESH_DIR, "block_cyl_case_templates"
)
BIRD_BLOCK_RECT_MESH_TEMP_DIR = os.path.join(
    BIRD_MESH_DIR, "block_rect_mesh_templates"
)
BIRD_BLOCK_RECT_CASE_TEMP_DIR = os.path.join(
    BIRD_MESH_DIR, "block_rect_case_templates"
)
BIRD_STIRRED_TANK_MESH_TEMP_DIR = os.path.join(
    BIRD_MESH_DIR, "stirred_tank_mesh_templates"
)
BIRD_STIRRED_TANK_CASE_TEMP_DIR = os.path.join(
    BIRD_MESH_DIR, "stirred_tank_case_templates"
)
BIRD_PRE_PATCH_TEMP_DIR = os.path.join(
    BIRD_PRE_DIR, "stl_patch", "bc_patch_mesh_template"
)
BIRD_PRE_DYNMIX_TEMP_DIR = os.path.join(
    BIRD_PRE_DIR, "dynamic_mixer", "mixing_template"
)
BIRD_EARLY_PRED_DATA_DIR = os.path.join(BIRD_POST_DIR, "data_early")
BIRD_KLA_DATA_DIR = os.path.join(BIRD_POST_DIR, "data_kla")
BIRD_CASE_GEN_DATA_DIR = os.path.join(BIRD_PRE_DIR, "data_case_gen")
BIRD_INV_DIR = os.path.join(BIRD_DIR, "inverse_modeling")

from bird.logging_config import setup_logging

setup_logging(level="INFO")
# setup_logging(level="DEBUG")



import argparse
import os
import sys

import numpy as np
from prettyPlot.plotting import plt, pretty_labels

from bird.postprocess.post_quantities import *
from bird.utilities.ofio import *

parser = argparse.ArgumentParser(description="Convergence of GH")
parser.add_argument(
    "-cn",
    "--case_name",
    type=str,
    metavar="",
    required=True,
    help="Case name",
)
parser.add_argument(
    "-df",
    "--data_folder",
    type=str,
    metavar="",
    required=False,
    help="data folder name",
    default="data",
)

args, unknown = parser.parse_known_args()


case_root = "."  # "../"
case_name = args.case_name  # "12_hole_sparger_snappyRefine_700rpm_opt_coeff"
case_path = "."
dataFolder = args.data_folder

if os.path.isfile(os.path.join(dataFolder, case_name, "conv.npz")):
    sys.exit("WARNING: History already created, Skipping")

time_float_sorted, time_str_sorted = getCaseTimes(case_path, remove_zero=True)
cellCentres = readMesh(os.path.join(case_path, f"meshCellCentres_0.obj"))
nCells = len(cellCentres)


co2_history = np.zeros(len(time_str_sorted))
c_co2_history = np.zeros(len(time_str_sorted))
h2_history = np.zeros(len(time_str_sorted))
c_h2_history = np.zeros(len(time_str_sorted))
gh_history = np.zeros(len(time_str_sorted))
liqvol_history = np.zeros(len(time_str_sorted))
print(f"case_path = {case_path}")
field_dict = {}
for itime, time in enumerate(time_float_sorted):
    time_folder = time_str_sorted[itime]
    print(f"\tTime : {time_folder}")
    if not field_dict == {}:
        new_field_dict = {}
        if "V" in field_dict:
            new_field_dict["V"] = field_dict["V"]
        field_dict = new_field_dict
    gh_history[itime], field_dict = compute_gas_holdup(
        case_path,
        time_str_sorted[itime],
        nCells,
        volume_time="0",
        field_dict=field_dict,
    )
    co2_history[itime], field_dict = compute_ave_y_liq(
        case_path,
        time_str_sorted[itime],
        nCells,
        volume_time="0",
        spec_name="CO2",
        field_dict=field_dict,
    )
    h2_history[itime], field_dict = compute_ave_y_liq(
        case_path,
        time_str_sorted[itime],
        nCells,
        volume_time="0",
        spec_name="H2",
        field_dict=field_dict,
    )
    c_co2_history[itime], field_dict = compute_ave_conc_liq(
        case_path,
        time_str_sorted[itime],
        nCells,
        volume_time="0",
        spec_name="CO2",
        mol_weight=0.04401,
        field_dict=field_dict,
    )
    c_h2_history[itime], field_dict = compute_ave_conc_liq(
        case_path,
        time_str_sorted[itime],
        nCells,
        volume_time="0",
        spec_name="H2",
        mol_weight=0.002016,
        field_dict=field_dict,
    )

os.makedirs(dataFolder, exist_ok=True)
os.makedirs(os.path.join(dataFolder, case_name), exist_ok=True)
np.savez(
    os.path.join(dataFolder, case_name, "conv.npz"),
    time=np.array(time_float_sorted),
    gh=gh_history,
    co2=co2_history,
    h2=h2_history,
    c_h2=c_h2_history,
    c_co2=c_co2_history,
)



# Import the relevant IO functions
from bird.utilities.ofio import *

# Read cell centers
cell_centers = readMesh("meshCellCentres_1.obj")
print("cell centers shape = ", cell_centers.shape)

# Read relevant fields at time 80
co2_gas = readOF("80/CO2.gas")
alpha_gas = readOF("80/alpha.gas")
u_liq = readOF("80/U.liquid")
print("cell CO2 gas shape = ", co2_gas["field"].shape)
print("cell alpha gas shape = ", alpha_gas["field"].shape)
print("cell u liq shape = ", u_liq["field"].shape)

# Compute conditional average of co2_gas and alpha_gas over y
from bird.utilities.mathtools import conditional_average

y_co2_gas_cond, co2_gas_cond = conditional_average(
    cell_centers[:, 1], co2_gas["field"], nbins=32
)
y_alpha_gas_cond, alpha_gas_cond = conditional_average(
    cell_centers[:, 1], alpha_gas["field"], nbins=32
)

# Plot
from prettyPlot.plotting import *

fig = plt.figure()
plt.plot(y_co2_gas_cond, co2_gas_cond, color="k", label=r"$Y_{CO_2}$ [-]")
plt.plot(
    y_alpha_gas_cond, alpha_gas_cond, color="b", label=r"$\alpha_{g}$ [-]"
)
pretty_labels("Y [m]", "", fontsize=20, grid=False, fontname="Times")
pretty_legend(fontname="Times")
plt.show()

# Compute reactor quantities
from bird.postprocess.post_quantities import *

kwargs = {"case_folder": ".", "time_folder": "80"}
gh, field_dict = compute_gas_holdup(
    volume_time="1", field_dict={"cell_centers": cell_centers}, **kwargs
)
print("fields stored = ", list(field_dict.keys()))
print(f"Gas Holdup = {gh:.4g}")
sup_vel, field_dict = compute_superficial_velocity(
    volume_time="1", field_dict=field_dict, **kwargs
)
print("fields stored = ", list(field_dict.keys()))
print(f"Superficial velocity = {sup_vel:.4g} m/s")
y_ave_co2, field_dict = compute_ave_y_liq(
    volume_time="1", spec_name="CO2", field_dict=field_dict, **kwargs
)
print("fields stored = ", list(field_dict.keys()))
print(f"Reactor averaged YCO2 = {y_ave_co2:.4g}")
c_ave_co2, field_dict = compute_ave_conc_liq(
    volume_time="1",
    spec_name="CO2",
    mol_weight=0.04401,
    rho_val=1000,
    field_dict=field_dict,
    **kwargs,
)
print("fields stored = ", list(field_dict.keys()))
print(f"Reactor averaged [CO2] = {c_ave_co2:.4g} mol/m3")
diam, field_dict = compute_ave_bubble_diam(
    volume_time="1", field_dict=field_dict, **kwargs
)
print("fields stored = ", list(field_dict.keys()))
print(f"Reactor averaged bubble diameter = {diam:.4g} m")



import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "bird", "requirements.txt")) as f:
    install_requires = f.readlines()

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "bird", "version.py"), encoding="utf-8") as f:
    version = f.read()
version = version.split("=")[-1].strip().strip('"').strip("'")

setup(
    name="nrel-bird",
    version=version,
    description="Bio Reactor Design (BiRD): a toolbox to simulate and analyze different designs of bioreactors in OpenFOAM",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/NREL/BioReactorDesign",
    author="Malik Hassanaly",
    license="BSD 3-Clause",
    package_dir={"bird": "bird"},
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.13",
    ],
    package_data={
        "": [
            "*requirements.txt",
            "*.yaml",
        ]
    },
    extras_require={
        "calibration": [
            "joblib",
            "tensorflow",
            "scikit-learn",
            "tf2jax",
        ],
        "optim": [
            "optuna",
            "pandas",
        ],
    },
    project_urls={
        "Documentation": "https://nrel.github.io/BioReactorDesign/",
        "Repository": "https://github.com/NREL/BioReactorDesign",
    },
    include_package_data=True,
    python_requires=">=3.10",
    install_requires=install_requires,
)



import logging
import os
import pickle

from prettyPlot.plotting import plt

from bird.postprocess.post_quantities import *
from bird.utilities.mathtools import *
from bird.utilities.ofio import *

logger = logging.getLogger(__name__)


def compute_cond_mean(
    case_path,
    vert_ind,
    field_name_list,
    window_ave,
    n_bins=32,
    diff_val_list=[],
    diff_name_list=[],
):
    time_float_sorted, time_str_sorted = getCaseTimes(case_path)
    mesh_time_str = getMeshTime(case_path)
    cellCentres = readMesh(
        os.path.join(case_path, f"meshCellCentres_{mesh_time_str}.obj")
    )
    nCells = len(cellCentres)
    assert len(diff_val_list) == len(diff_name_list)
    window_ave = min(window_ave, len(time_str_sorted))

    fields_cond = {}
    fields_cond_tmp = {}
    for name in field_name_list:
        fields_cond[name] = {}
        fields_cond_tmp[name] = {}

    logger.info(f"Case : {case_path}")

    for i_ave in range(window_ave):
        time_folder = time_str_sorted[-i_ave - 1]
        logger.debug(f"\tReading Time : {time_folder}")
        field_file = []
        for field_name in field_name_list:
            field_file.append(os.path.join(case_path, time_folder, field_name))

        for filename, name in zip(field_file, field_name_list):
            val_dict = {}
            field_tmp = readOFScal(filename, nCells)["field"]
            vert_axis, field_cond_tmp = conditional_average(
                cellCentres[:, vert_ind], field_tmp, nbins=n_bins
            )
            if i_ave == 0:
                fields_cond[name]["val"] = field_cond_tmp / window_ave
                fields_cond[name]["vert"] = vert_axis
            else:
                fields_cond[name]["val"] += field_cond_tmp / window_ave

    return fields_cond


def save_cond(filename, fields_cond):
    with open(filename, "wb") as f:
        pickle.dump(fields_cond, f)


def sequencePlot(
    cond,
    folder_names,
    field_name,
    case_names=[],
    symbList=["-", "-d", "-^", "-.", "-s", "-o", "-+"],
):
    if not len(case_names) == len(folder_names):
        case_names = [f"test{i}" for i in range(len(folder_names))]
    if len(case_names) > len(symbList):
        error_msg = f"too many cases ({len(case_names)}), reduce number of case to {len(symbList)} or add symbols"
        logger.error(error_msg)
        raise IndexError(error_msg)

    for ic, (case_name, folder_name) in enumerate(
        zip(case_names, folder_names)
    ):
        label = ""
        if ic == 0:
            label = case_names[ic]
        plt.plot(
            cond[folder_name][field_name]["val"],
            cond[folder_name][field_name]["vert"],
            symbList[ic],
            markersize=10,
            markevery=10,
            linewidth=3,
            color="k",
            label=label,
        )
    if field_name == "alpha.gas":
        plot_name = "gasHoldup"
    else:
        plot_name = field_name
    return plot_name



import json
import os
import pickle as pkl

import matplotlib as mpl
import numpy as np
from prettyPlot.plotting import *
from scipy.optimize import curve_fit


def get_sim_folds(path):
    folds = os.listdir(path)
    sim_folds = []
    for fold in folds:
        if fold.startswith("loop"):
            sim_folds.append(fold)
    return sim_folds


def func(t, cstar, kla):
    t = t
    t0 = 0
    c0 = 0
    return (cstar - c0) * (1 - np.exp(-kla * (t - t0))) + c0


def get_vl(verb=False):
    filename = os.path.join("constant", "globalVars")
    with open(filename, "r+") as f:
        lines = f.readlines()
    for line in lines:
        if line.startswith("liqVol"):
            vol = float(line.split()[-1][:-1])
            break
    if verb:
        print(f"Read liqVol = {vol}m3")
    return vol


def get_vvm(verb=False):
    filename = os.path.join("constant", "globalVars")
    with open(filename, "r+") as f:
        lines = f.readlines()
    for line in lines:
        if line.startswith("VVM"):
            vvm = float(line.split()[-1][:-1])
            break
    if verb:
        print(f"Read VVM = {vvm} [-]")
    return vvm


def get_As(verb=False):
    filename = os.path.join("constant", "globalVars")
    with open(filename, "r+") as f:
        lines = f.readlines()
    for line in lines:
        if line.startswith("inletA"):
            As = float(line.split()[-1][:-1])
            break
    if verb:
        print(f"Read As = {As}m2")
    return As


def get_pmix(verb=False):
    with open("system/mixers.json", "r+") as f:
        data = json.load(f)
    mixer_list = data["mixers"]
    pmix = 0
    for mix in mixer_list:
        pmix += mix["power"] / 1000
    if verb:
        print(f"Read Mixing power = {pmix}kW")
    return pmix


def get_lh(verb=False):
    filename = os.path.join("system", "setFieldsDict")
    with open(filename, "r+") as f:
        lines = f.readlines()
    for line in lines:
        if "box (-1.0 -1.0 -1.0)" in line:
            height = float(line.split("(")[2].split()[1])
            break
    if verb:
        print(f"Read Height = {height}m")
    return height


def get_pinj(vvm, Vl, As, lh):
    rhog = 1.25  # kg /m3
    Vg = Vl * vvm / (60 * As * 1)  # m/s
    Ptank = 101325  # Pa
    # Ptank = 0 # Pa
    rhoL = 1000  # kg / m3
    Pl = 101325 + rhoL * 9.8 * lh  # Pa
    # W
    P1 = rhog * As * Vg**3
    # W
    P2 = (Pl - Ptank) * As * Vg
    # kg /s
    MF = rhog * Vg * As
    # kwh / kg
    e_m = (P1 + P2) / (3600 * 1000 * MF)

    # returns kW
    return (P1 + P2) * 1e-3


def get_qoi(kla_co2, cs_co2, kla_h2, cs_h2, verb=False):
    vvm = get_vvm(verb)
    As = get_As(verb)
    V_l = get_vl(verb)
    liqh = get_lh(verb)
    P_inj = get_pinj(vvm, V_l, As, liqh)
    P_mix = get_pmix(verb)

    qoi_co2 = kla_co2 * cs_co2 * V_l * 0.04401 / (P_mix / 3600 + P_inj / 3600)
    qoi_h2 = kla_h2 * cs_h2 * V_l * 0.002016 / (P_mix / 3600 + P_inj / 3600)
    return qoi_co2 * qoi_h2


def get_qoi_uq(kla_co2, cs_co2, kla_h2, cs_h2):
    qoi = []
    for i in range(len(kla_co2)):
        if i == 0:
            verb = True
        else:
            verb = False
        qoi.append(get_qoi(kla_co2[i], cs_co2[i], kla_h2[i], cs_h2[i], verb))
    qoi = np.array(qoi)
    return np.mean(qoi), np.std(qoi)


os.makedirs("Figures", exist_ok=True)

dataFolder = "data"
fold = "local"

nuq = 100
mean_cstar_co2 = np.random.uniform(14, 16.9, nuq)
mean_cstar_h2 = np.random.uniform(1.04, 1.19, nuq)


tmp_cs_h2 = []
tmp_cs_co2 = []
tmp_kla_h2 = []
tmp_kla_co2 = []
cs_co2 = mean_cstar_co2
cs_h2 = mean_cstar_h2

a = np.load(os.path.join(dataFolder, fold, "conv.npz"))
endindex = -1
if (
    "c_h2" in a
    and "c_co2" in a
    and len(a["time"][:endindex] > 0)
    and (a["time"][:endindex][-1] > 95)
):
    for i in range(nuq):
        fitparamsH2, _ = curve_fit(
            func,
            np.array(a["time"][:endindex]),
            np.array(a["c_h2"][:endindex]),
            bounds=[(cs_h2[i] - 1e-6, 0), (cs_h2[i] + 1e-6, 1)],
        )
        fitparamsCO2, _ = curve_fit(
            func,
            np.array(a["time"][:endindex]),
            np.array(a["c_co2"][:endindex]),
            bounds=[(cs_co2[i] - 1e-6, 0), (cs_co2[i] + 1e-6, 1)],
        )
        tmp_kla_co2.append(fitparamsCO2[1])
        tmp_kla_h2.append(fitparamsH2[1])
        tmp_cs_h2.append(cs_h2[i])
        tmp_cs_co2.append(cs_co2[i])

qoi_m, qoi_s = get_qoi_uq(tmp_kla_co2, tmp_cs_co2, tmp_kla_h2, tmp_cs_h2)


with open("qoi.txt", "w+") as f:
    f.write(f"{qoi_m},{qoi_s}\n")
